# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import logging
import signal
import socket
import urlparse
import threading
import collections

from lmi.shell.compat import *

from lmi.shell.LMIBaseObject import LMIWrapperBaseObject
from lmi.shell.LMIShellConfig import LMIShellConfig
from lmi.shell.LMIObjectFactory import LMIObjectFactory
from lmi.shell.LMIFormatter import LMIMethodFormatter
from lmi.shell.LMIFormatter import LMIMofFormatter
from lmi.shell.LMIReturnValue import LMIReturnValue
from lmi.shell.LMIConstantValues import LMIConstantValuesParamProp
from lmi.shell.LMIConstantValues import LMIConstantValuesMethodReturnType
from lmi.shell.LMIIndicationListener import LMIIndicationListener

from lmi.shell.LMIJob import lmi_is_job_finished
from lmi.shell.LMIJob import lmi_is_job_completed
from lmi.shell.LMIJob import lmi_is_job_terminated
from lmi.shell.LMIJob import lmi_is_job_killed
from lmi.shell.LMIJob import lmi_is_job_exception
from lmi.shell.LMIJob import JOB_NOT_FINISHED
from lmi.shell.LMIJob import JOB_FINISH_DELAYED
from lmi.shell.LMIJob import JOB_FINISH_EARLY

from lmi.shell.LMIUtil import LMIPassByRef
from lmi.shell.LMIUtil import lmi_get_use_exceptions
from lmi.shell.LMIUtil import lmi_raise_or_dump_exception
from lmi.shell.LMIUtil import lmi_transform_to_cim_param
from lmi.shell.LMIUtil import lmi_transform_to_lmi
from lmi.shell.LMIUtil import lmi_wrap_cim_class

from lmi.shell.LMIExceptions import *

logger = logging.getLogger(__name__)


class LMISignalHelper(object):
    """
    Helper class, which takes care of signal (de)registration and handling.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(LMISignalHelper, cls).__new__(cls)
            LMISignalHelper.reset(cls._instance)
        return cls._instance

    def reset(self):
        """
        Resets the single instance into default state.
        """
        self._handler_sigint = None
        self._handler_sigterm = None
        self._signal_handled = False
        self._instance._callbacks = collections.OrderedDict()

    def signal_attach(self):
        """
        Registers *SIGINT* and *SIGTERM* signals to local handler in which, the
        flags for each signal are modified, if such signal is caught.
        """
        self._signal_handled = False
        self._handler_sigint = signal.signal(
            signal.SIGINT,  LMISignalHelper.__signal_handler)
        self._handler_sigterm = signal.signal(
            signal.SIGTERM, LMISignalHelper.__signal_handler)

    def signal_detach(self):
        """
        Unregisters *SIGINT* and *SIGTERM* handler and removes all the attached
        callbacks.
        """
        signal.signal(signal.SIGINT,  self._handler_sigint)
        signal.signal(signal.SIGTERM, self._handler_sigterm)

    def signal_handled(self):
        """
        :returns: True, if any of *SIGINT* or *SIGTERM* has been caught; \
            False otherwise
        """
        return self._signal_handled

    def callback_attach(self, cb_name, cb):
        """
        Registers a callback, which will be called when a *SIGINT* or *SIGTERM*
        is caught.

        :param string cb_name: callback name
        :param cb: callable object, which takes zero arguments
        """
        self._callbacks[cb_name] = cb

    def callback_detach(self, cb_name):
        """
        Removes a callback from the callback dictionary.

        :param string cb_name: callback name
        """
        self._callbacks.pop(cb_name)

    @staticmethod
    def __signal_handler(signo, frame):
        """
        Signal handler, which is called, when *SIGINT* and *SIGTERM* are sent
        to the LMIShell.

        :param int signo: signal number
        :param frame: -- stack frame

        **NOTE:** see help(signal)
        """
        if signo in (signal.SIGINT, signal.SIGTERM):
            LMISignalHelper._instance._signal_handled = True
        for cb in LMISignalHelper._instance._callbacks.values():
            cb()


class LMIMethod(LMIWrapperBaseObject):
    """
    LMI wrapper class representing :py:class:`wbem.CIMMethod`.

    :param LMIConnection conn: connection object
    :param LMIInstance(Name) lmi_instance: :py:class:`.LMIInstance` or
        :py:class:`.LMIInstanceName` object, on which the method call will be
        issued
    :param string method_name: method name
    """
    # 15 seconds sleep timeout for main waiting thread
    _COND_WAIT_TIME = 15
    # Wake count of main thread, when the GetInstance is performed to check,
    # if the job object is present. Prevents infinite waiting for indication
    # delivery. Maximum waiting time, before the GetInstance for job object
    # will be called is: _COND_WAIT_TIME * _COND_WAIT_WAKE_CNT
    _COND_WAIT_WAKE_CNT = 4
    # Default tcp port, where the indications will be delivered.
    # TODO: create a configuration option for the port
    _INDICATION_DESTINATION_PORT = 10240
    # Default number retries when bind fails.
    _INDICATION_BIND_TRIES = 10
    # Job classes, which can be used for synchro method calls
    # TODO: create a configuration option for the static filters' classnames
    _INDICATION_JOB_CLASSNAMES = (
        "LMI_SELinuxJob",
        "LMI_StorageJob",
        "LMI_SoftwareInstallationJob",
        "LMI_SoftwareVerificationJob",
        "LMI_NetworkJob")
    # Default namespace where the indication subscriptions, used for
    # synchronous method calls, will be registered.
    _INDICATION_NAMESPACE = "root/interop"
    # When performing a synchronous method call and using the polling method to
    # get a job object status, the sleep time between 2 polls doubles if it is
    # less than _POLLING_ADAPT_MAX_WAITING_TIME.
    _POLLING_ADAPT_MAX_WAITING_TIME = 32

    def __init__(self, conn, lmi_instance, method_name):
        super(LMIMethod, self).__init__(conn)
        if not isinstance(lmi_instance, (
                LMIObjectFactory().LMIInstance,
                LMIObjectFactory().LMIInstanceName)):
            raise TypeError("lmi_instance must be LMIInstance(Name) type")
        self._lmi_instance = lmi_instance
        self._sync_method = False
        self._method = None
        self._method_name = method_name
        self._valuemap_parameters_list = []

        is_sync = method_name.startswith("Sync")

        if not conn.is_wsman():
            if is_sync:
                # Store the synchronous flag and trim the method name
                self._sync_method = True
                self._method_name = method_name[4:]

            # We need to have CIMClass with qualifiers. Fetch full class.
            if isinstance(self._lmi_instance, LMIObjectFactory().LMIInstance):
                # LMIInstance
                lmi_class = self._lmi_instance._lmi_class
            else:
                # LMIInstanceName
                lmi_class = lmi_wrap_cim_class(
                    self._conn,
                    self._lmi_instance.classname,
                    self._lmi_instance.namespace)
            lmi_class.fetch(True)

            self._method = lmi_class._cim_class.methods[self._method_name]

            # Store the constant values as a list. This can consume some time, if
            # computed on demand.
            self._valuemap_parameters_list = [
                k for k, v in self._method.parameters.iteritems()
                if "ValueMap" in v.qualifiers]

            # For simplicity, we add return value constants to the same list
            if "ValueMap" in self._method.qualifiers:
                self._valuemap_parameters_list.append(self._method_name)
        elif conn.is_wsman() and is_sync:
            # XXX: We can't perform synchro calls using WSMAN
            errorstr = "Can't perform synchronous method calls using WSMAN"
            raise ValueError(errorstr)

    def __return_synchro_method_call(self, job_inst, job_refresh=True):
        """
        Returns a :py:class:`.LMIReturnValue` object with Job output parameters
        set.

        :param LMIInstance job_inst: job returned from a synchronous method
            call
        :param bool job_refresh: flag, which indicates, if the ``job_inst``
            needs to be refreshed
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to Job
            output parameters
        """
        def get_rval_errorstr(err_list):
            for err in err_list:
                if err.CIMStatusCode is not None:
                    return err.CIMStatusCode, err.Message
            # There is no CIM_Error instance with CIMStatusCode set. In such
            # case, we set rval to CIM_ERR_FAILED and for error message we use
            # the first instance from err_list.
            message = err_list[0].Message if err_list else "Failed"
            return wbem.CIM_ERR_FAILED, message

        # Adjust return value from the job object.
        if job_refresh:
            job_inst.refresh()
        rval = None
        rparams = wbem.NocaseDict()
        if job_inst.JobOutParameters is not None:
            props = job_inst.JobOutParameters.properties
            rparams = wbem.NocaseDict()
            for name, prop in props.iteritems():
                # Return just those properties of given class that are listed
                # as output parameters of particular method.
                if self._method.parameters.has_key(name) \
                        and (self._method.parameters[name].qualifiers is None \
                            or self._method.parameters[name].qualifiers.get(
                                'Out', False)):
                    rparams[name] = prop.value
                elif name == '__ReturnValue':
                    rval = prop.value
        errorstr = ""

        # Is job in exception state? If so, adjust corresponding error string
        # from job.GetError() instance
        if lmi_is_job_exception(job_inst):
            try:
                refreshed, _, errorstr = job_inst.refresh()
            except CIMError, e:
                logger.debug(
                    "Sync%s: Job instance can not be refreshed; %s" %
                    (self._method_name, str(e)))
                job_exception.value = e
            if not refreshed:
                logger.debug(
                    "Sync%s: Job instance can not be refreshed; %s" %
                    (self._method_name, errorstr))
                raise LMISynchroMethodCallError(errorstr)
            exc_rval, exc_rparams, exc_errorstr = job_inst.GetErrors()
            err_list = exc_rparams.get("errors", None)
            if not err_list:
                logger.debug(
                    "Sync%s: Can not retrieve Error instances; %s" %
                    (self._method_name, exc_errorstr))
                raise LMISynchroMethodCallError(
                    "Could not get Job error message")

            rval, errorstr = get_rval_errorstr(err_list)
            if lmi_get_use_exceptions():
                raise CIMError(rval, errorstr)

        return LMIReturnValue(rval=rval, rparams=rparams, errorstr=errorstr)

    def __handle_synchro_method_call_indication(self, job_inst):
        """
        Handles a synchronous call for asynchronous methods returning a job
        object. This method uses static filters installed by each OpenLMI
        provider, which is capable of using jobs.

        :param LMIInstance job_inst: job object returned from a synchronous
            method call
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to 0 and
            ``rparams`` set to job's output parameters
        :raises: :py:exc:`.LMIIndicationListenerError`,
            :py:exc:`.LMISynchroMethodCallError`,
            :py:exc:`.LMISynchroMethodCallFilterError`

        **NOTE:** Static filters' names need to be in format
            "``LMI:<job_class_name>:Changed``"

        """
        def handle_job(ind, cond, job_finished, job_exception):
            """
            :py:class:`.LMIListener` handler for synchronous method call, which
            uses indication means of waiting for the job. This function is
            called, when a job changes its state.

            :param threading.Condition cond: condition object used for thread
                synchronization
            :param LMIPassByRef job_finished: used for synchronization, whether
                the job has finished
            :param LMIPassByRef job_exception: contains an exception object, if
                any exception was raised
            """
            cond.acquire()
            try:
                src_inst = ind["SourceInstance"]
                if lmi_is_job_finished(src_inst):
                    # Job has just finished
                    job_finished.value = JOB_FINISH_DELAYED
            except Exception, e:
                # Notify main thread, we are not able to work with such
                # objects.
                job_finished.value = JOB_FINISH_DELAYED
                job_exception.value = e
            finally:
                # XXX: Let's be defensive, always notify+release main thread.
                cond.notify()
                cond.release()

        cond = threading.Condition()
        job_finished = LMIPassByRef(JOB_NOT_FINISHED)
        job_exception = LMIPassByRef(None)
        # There needs to be a pattern of at least 8 "X" in a row at the end of
        # the indication_name.
        indication_name = "synchro-method-call-XXXXXXXX"

        # Start the indication listener
        listener = LMIIndicationListener(
            "0.0.0.0",
            self._INDICATION_DESTINATION_PORT,
            LMIShellConfig().cert_file,
            LMIShellConfig().key_file)
        indication_name = listener.add_handler(
            indication_name, handle_job, cond, job_finished, job_exception)
        listener.start(self._INDICATION_BIND_TRIES)
        logger.debug(
            "Choosing port for Sync%s() indications: %d" %
            (self._method_name, listener.port))

        # Search for necessary static filter
        filter_name = "LMI:%s:Changed" % job_inst.classname
        cim_filters, _, _ = self._conn.client.get_instances(
            "CIM_IndicationFilter", LMIMethod._INDICATION_NAMESPACE,
            {"Name": filter_name})
        if not cim_filters:
            listener.stop()
            errorstr = "Can not find proper CIM_IndicationFilter"
            raise LMISynchroMethodCallFilterError(errorstr)
        cim_filter = cim_filters[0]

        netloc = urlparse.urlparse(self._conn.uri).netloc
        if not netloc:
            listener.stop()
            errorstr = "Can not determine netloc from client's uri"
            raise LMISynchroMethodCallError(errorstr)
        netloc = netloc.split(":")[0]
        # NOTE: This will work only on a local area network. Complicated
        # networks may require additional configuration to make this work. See
        # LMIMethod() and PreferPolling.
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect((netloc, listener.port))
        except socket.gaierror, e:
            listener.stop()
            errorstr = "Can not determine IP address of this machine"
            raise LMISynchroMethodCallError(errorstr)
        destination = s.getsockname()[0]
        s.close()

        # Create handler object
        cim_handler_props = {
            "Name": indication_name,
            "Destination": "%s://%s:%d/CIMListener/%s" % (
                "https" if listener.uses_ssl else "http",
                destination, listener.port, indication_name),
        }
        cim_handler, _, _ = self._conn.client.create_instance(
            "CIM_IndicationHandlerCIMXML",
            LMIMethod._INDICATION_NAMESPACE,
            self._conn.hostname,
            cim_handler_props)
        if not cim_handler:
            listener.stop()
            errorstr = "Can not create CIM_IndicationHandlerCIMXML object"
            raise LMISynchroMethodCallError(errorstr)

        # Create indication subscription object
        cim_subscription_props = {
            "Filter": cim_filter.path,
            "Handler": cim_handler.path
        }
        cim_subscription, _, _ = self._conn.client.create_instance(
            "CIM_IndicationSubscription",
            LMIMethod._INDICATION_NAMESPACE,
            self._conn.hostname,
            cim_subscription_props)
        if not cim_subscription:
            self._conn.client.delete_instance(cim_handler)
            listener.stop()
            errorstr = "Can not create CIM_IndicationSubscription object"
            raise LMISynchroMethodCallError(errorstr)

        # Check, if the job is not already in finished state,
        # while we were subscribing for the indications
        job_inst.refresh()
        if lmi_is_job_finished(job_inst):
            job_finished.value = JOB_FINISH_EARLY

        # Register signal callback for SIGINT, SIGTERM with callback,
        # which awakes waiting thread for immediate return.
        LMISignalHelper().callback_attach(
            "indication", lambda: LMIMethod.__wake(cond))
        LMISignalHelper().signal_attach()

        # Wait for the job to finish
        wake_cnt = 0
        cond.acquire()
        while not LMISignalHelper().signal_handled() and \
                not job_finished.value and \
                not lmi_is_job_finished(job_inst):
            cond.wait(LMIMethod._COND_WAIT_TIME)
            wake_cnt += 1
            # XXX: threading.Condition.wait() does not inform about timeout or
            # being awaken by notify call. There is a counting to 4 sleep
            # cycles before we actually check for job status manually. This
            # number can be increased, so we rely more on indications, rather
            # then on manual polling.
            if wake_cnt >= LMIMethod._COND_WAIT_WAKE_CNT and \
                    not job_finished.value:
                wake_cnt = 0
                try:
                    refreshed, _, errorstr = job_inst.refresh()
                except CIMError, e:
                    logger.debug(
                        "Sync%s: Job instance can not be refreshed; %s" %
                        (self._method_name, str(e)))
                    job_exception.value = e
                    break
                if not refreshed:
                    logger.debug(
                        "Sync%s: Job instance can not be refreshed; %s" %
                        (self._method_name, errorstr))
                    job_exception.value = LMISynchroMethodCallError(errorstr)
                    break

        # Unregister signal handler
        LMISignalHelper().signal_detach()
        LMISignalHelper().callback_detach("indication")

        cond.release()

        # Cleanup
        listener.stop()
        self._conn.client.delete_instance(cim_subscription.path)
        self._conn.client.delete_instance(cim_handler.path)
        if job_exception.value:
            raise job_exception.value
        if LMISignalHelper().signal_handled() and not job_finished.value:
            # We got SIGINT or SIGTERM, when waiting for the job, cancelling
            # the job
            logger.warn("Cancelling a job '%s'" % job_inst.Name)
            rstate = job_inst.RequestStateChange.RequestedStateValues.Terminate
            job_inst.RequestStateChange(RequestedState=rstate)
            return LMIReturnValue(rval=None)

        # Return the job return values, refresh the job_inst object, if we got
        # notified about the job finish state by indication.
        return self.__return_synchro_method_call(
            job_inst, job_finished.value == JOB_FINISH_DELAYED)

    def __handle_synchro_method_call_polling(self, job_inst):
        """
        Handles a synchronous call for asynchronous methods returning a job
        object.  This call uses polling method to wait for the job to finish.

        :param LMIInstance job_inst: job object
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to 0 and
            ``rparams`` set to job's output parameters.
        :raises: :py:exc:`.LMISynchroMethodCallError`
        """

        # Register signal callback for SIGINT, SIGTERM with callback,
        # which awakes waiting thread for immediate return.
        LMISignalHelper().callback_attach(
            "polling", lambda: LMIMethod.__wake(cond))
        LMISignalHelper().signal_attach()

        cond = threading.Condition()
        cond.acquire()

        job_exception = None

        try:
            sleep_time = 1
            while not LMISignalHelper().signal_handled() and \
                    not lmi_is_job_finished(job_inst):
                # Sleep, a bit longer in every iteration
                cond.wait(sleep_time)
                if sleep_time < LMIMethod._POLLING_ADAPT_MAX_WAITING_TIME:
                    sleep_time *= 2
                refreshed, _, errorstr = job_inst.refresh()
                if not refreshed:
                    logger.debug(
                        "Sync%s: Job instance can not be refreshed; %s" %
                        (self._method_name, errorstr))
                    job_exception = LMISynchroMethodCallError(errorstr)
                    break
        except CIMError, e:
            logger.debug(
                "Sync%s: Job instance can not be refreshed; %s" %
                (self._method_name, str(e)))
            job_exception = LMISynchroMethodCallError(e.message)
        finally:
            cond.release()

        # Unregister signal handler and callback
        LMISignalHelper().signal_detach()
        LMISignalHelper().callback_detach("polling")

        if LMISignalHelper().signal_handled() and \
                not lmi_is_job_finished(job_inst):
            # We got SIGINT or SIGTERM, when waiting for the job, cancelling
            # the job
            logger.warn("Cancelling a job '%s'" % job_inst.Name)
            rstate = job_inst.RequestStateChange.RequestedStateValues.Terminate
            job_inst.RequestStateChange(RequestedState=rstate)
            return LMIReturnValue(rval=None)

        if job_exception is not None:
            raise job_exception

        # Return the job return values. No need to refresh the job instance, we
        # already have a "fresh" one.
        return self.__return_synchro_method_call(job_inst, False)

    def __call__(self, method_args=None, **kwargs):
        """
        Perform a method call.

        Method arguments are preferably passed by dictionary (parameter :
        value). Using former means of passing arguments, by keyword arguments,
        to a method call works too.

        If performing a synchronous method call, passing PreferPolling can be
        used to select which method should be used -- either subscribing to an
        indication or polling method. This is available only, when talking to
        CIMOM via CIM-XML.

        :param dictionary method_args: method arguments
        :param dictionary kwargs: keyword method arguments (``method_args``
            preffered)

            * **RefreshInstance** (*bool*) flag, which tells the LMIShell,
              whether the instance should be refreshed after a method call.
              Default value is False.

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to
            method's return value; ``rparams`` set to method's return
            parameters; ``errorstr`` set method's error string in case of
            failure
        :raises: :py:exc:`.LMIUnknownParameterError`,
            :py:exc:`.LMIMethodCallError`
            :py:exc:`.LMISynchroMethodCallError`

        **Usage:** :ref:`instances_methods`.
        """
        synchro_method_polling = kwargs.pop("PreferPolling", False)
        refresh_requested = kwargs.pop("RefreshInstance", False)

        # Prepare method parameters
        if method_args is None:
            method_args = {}
        method_args.update(kwargs)
        if not self._conn.is_wsman():
            for param, value in method_args.iteritems():
                if param in self._method.parameters:
                    # Cast input parameters into acceptable CIM types
                    t = self._method.parameters[param].type
                    method_args[param] = lmi_transform_to_cim_param(t, value)
                else:
                    # NOTE: maybe we could check for wbem type and not to exit
                    # prematurely
                    errorstr = "Unknown parameter '%s' supplied for method '%s'" % \
                        (param, self._method.name)
                    lmi_raise_or_dump_exception(LMIUnknownParameterError(errorstr))
                    return LMIReturnValue(rval=-1, errorstr=errorstr)
        else:
            for param, value in method_args.iteritems():
                method_args[param] = str(value)

        # Call CIM method
        client = self._conn.client
        rval, call_rparams, call_errorstr = client.call_method(
            self._lmi_instance, self._method_name, **method_args)
        rval = lmi_transform_to_lmi(self._conn, rval)

        # Do we have any return parameters? If so, transform them to LMIShells'
        # types. If it is possible, perform synchronous method call.
        if call_rparams:
            call_rparams = lmi_transform_to_lmi(self._conn, call_rparams)
            if not call_rparams:
                # NOTE: this is wrong! What should we do?
                errorstr = "Could not perform CIM -> LMI object transformation"
                lmi_raise_or_dump_exception(LMIMethodCallError(errorstr))
                return LMIReturnValue(rval=rval, errorstr=errorstr)
            # Check if we can perform synchronous method call
            job = call_rparams.get("job", None)
            can_perform_sync_call = False
            if not self._conn.is_wsman() and job and \
                    job.classname in LMIMethod._INDICATION_JOB_CLASSNAMES:
                # We can perform synchro methods only using CIM-XML
                can_perform_sync_call = True

            # Perform synchronous method call
            if self._sync_method and can_perform_sync_call:
                # We work with LMIInstance object, which simplifies the code
                # with instance refreshing and stuff.
                job_inst = call_rparams["job"].to_instance()
                # At first, try to wait for the call to finish by subscribing
                # to an indication
                handled_by_indication = False
                if not synchro_method_polling:
                    try:
                        rval, call_rparams, call_errorstr = \
                            self.__handle_synchro_method_call_indication(
                                job_inst)
                        handled_by_indication = True
                    except CIMError, e:
                        lmi_raise_or_dump_exception(e)
                        return LMIReturnValue(rval=-1, errorstr=e.args[1])
                    except LMISynchroMethodCallError, e:
                        lmi_raise_or_dump_exception(e)
                        return LMIReturnValue(rval=-1, errorstr=e.message)
                    # Fall through, try to handle the synchro call by polling
                    except LMIHandlerNamePatternError, e:
                        handled_by_indication = False
                    except LMISynchroMethodCallFilterError, e:
                        handled_by_indication = False
                    except LMIIndicationListenerError, e:
                        handled_by_indication = False
                if not handled_by_indication:
                    # Executed, when LMIListener can not be started
                    try:
                        rval, call_rparams, call_errorstr = \
                            self.__handle_synchro_method_call_polling(job_inst)
                    except LMISynchroMethodCallError, e:
                        lmi_raise_or_dump_exception(e)
                        return LMIReturnValue(rval=-1, errorstr=e.message)
                call_rparams = lmi_transform_to_lmi(self._conn, call_rparams)

        # Refresh is requested
        if refresh_requested:
            # Refresh the instance, within which the method was called
            ref_rval, _, ref_errorstr = self.__refresh_instance()
            if not ref_rval:
                errorstr = "Could not update an LMI object after a method call"
                lmi_raise_or_dump_exception(LMIMethodCallError(ref_errorstr))
                return LMIReturnValue(rval=ref_rval, errorstr=ref_errorstr)

        return LMIReturnValue(
            rval=rval, rparams=call_rparams, errorstr=call_errorstr)

    def __getattr__(self, name):
        """
        Returns either a class member, or a constant value.

        :param string name: class member, or the constant value name
        """
        if not self._conn.is_wsman() and name.endswith("Values"):
            parameter_name = name[:-6]
            if parameter_name in self._method.parameters:
                return LMIConstantValuesParamProp(
                    self._method.parameters[parameter_name])
            elif parameter_name == self._method.name:
                return LMIConstantValuesMethodReturnType(self._method)
        raise AttributeError(name)

    @staticmethod
    def __wake(cond):
        """
        Helper function used for manual :py:attr:`threading.Condition` wakeup.

        :param threading.Condition cond: condition object
        """
        cond.acquire()
        cond.notify()
        cond.release()

    def __refresh_instance(self):
        """
        Refreshes nested :py:class:`.LMIInstance`.

        **NOTE:** This method refreshes only nested :py:class:`.LMIInstance`.
        Objects of this class may also contain :py:class:`.LMIInstanceName`; in
        this case, no refreshing is performed.

        :rtype: :py:class:`.LMIReturnValue`
        """
        if not isinstance(self._lmi_instance, LMIObjectFactory().LMIInstance):
            # There is nothing to refresh here.
            return LMIReturnValue(rval=True)
        return self._lmi_instance.refresh()

    def doc(self):
        """
        Prints out pretty verbose message with documentation for the class. If
        the LMIShell is run in a interactive mode, the output will be
        redirected to a pager set by environment variable :envvar:`PAGER`. If
        there is not :envvar:`PAGER` set, less or more will be used as a
        fall-back.
        """
        if self._conn.is_wsman():
            # WSMAN doesn't support reflextion, we can't call GetClass().
            lmi_get_logger().info("WSMAN client doesn't support GetClass()")
            return

        formatter = LMIMethodFormatter(self._method)
        formatter.fancy_format(self._conn.client.interactive)

    @property
    def return_type(self):
        """
        :returns: string of the method call's return type
        """
        if self._conn.is_wsman():
            # WSMAN doesn't support reflextion, we can't call GetClass().
            return "unknown"
        return self._method.return_type

    def tomof(self):
        """
        Prints out a message with MOF representation of
        :py:class:`wbem.CIMMethod`.  If the LMIShell is run in a interactive
        mode, the output will be redirected to a pager set by environment
        variable :envvar:`PAGER`. If there is not :envvar:`PAGER` set, less or
        more will be used as a fall-back.
        """
        if self._conn.is_wsman():
            # WSMAN doesn't support reflextion, we can't call GetClass().
            lmi_get_logger().info("WSMAN client doesn't support GetClass()")
            return

        formatter = LMIMofFormatter(self._method)
        formatter.fancy_format(self._conn.client.interactive)

    def valuemap_parameters(self):
        """
        :returns: list of strings of the constant names
        """
        return self._valuemap_parameters_list

    def print_valuemap_parameters(self):
        """
        Prints out the list of strings of constant names.
        """
        for i in self._valuemap_parameters_list:
            sys.stdout.write("%s\n" % i)

    def parameters(self):
        """
        :returns: list of strings of :py:class:`wbem.CIMMethod`'s parameters
        """
        if self._conn.is_wsman():
            # WSMAN doesn't support reflextion, we can't call GetClass().
            return []

        return self._method.parameters

    def print_parameters(self):
        """
        Prints out :py:class:`wbem.CIMMethod`'s parameters.
        """
        if self._conn.is_wsman():
            # WSMAN doesn't support reflextion, we can't call GetClass().
            return

        for param, value in self._method.parameters.iteritems():
            sys.stdout.write(
                "%s %s%s\n" %
                (value.type, param, "[]" if value.is_array else ""))

    @property
    def wrapped_object(self):
        """
        :returns: wrapped :py:class:`wbem.CIMmethod` object
        """
        return self._method



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from functools import wraps

from lmi.shell.compat import *

from lmi.shell.LMIReturnValue import LMIReturnValue
from lmi.shell.LMIUtil import lmi_wrap_cim_class
from lmi.shell.LMIUtil import lmi_raise_or_dump_exception

from lmi.shell.LMIExceptions import CIMError
from lmi.shell.LMIExceptions import ConnectionError
from lmi.shell.LMIExceptions import LMIDeletedObjectError


class lmi_return_expr_if_fail(object):
    """
    Decorator, which calls a specified expression and returns its return value
    instead of calling the decorated method, if provided test expression is
    False; otherwise a method is called.

    :param expr_test: expression which determines, if to execute a return value
        expression
    :param expr_ret: expression, which is called, if the ``expr_test`` returns
        False
    :param expr_ret_args: ``expr_ret`` position arguments
    :param expr_ret_kwargs: ``expr_ret`` keyword arguments
    :param bool Self: flag, which specifies, if to pass ``self`` variable to
        the ``expr_ret``, if ``expr_test`` failed

    Example of usage:

    .. code-block:: python

        class Foo:
            def __init__(self, member):
                self._member = member

            def failed(self):
                print "expression failed"
                return False

            # NOTE: the self parameter to the method call needs to be passed
            # via expr_ret_args, therefore, there is a dummy lambda obj: obj,
            # which is basically self variable.
            @lmi_return_expr_if_fail(lambda obj: obj._member, failed,
                                     lambda obj: obj)
            def some_method(self):
                print "some_method called"
                return True

        f = Foo(None)
        f.some_method() == False

        f = Foo(True)
        f.some_method() == True
    """
    def __init__(self, expr_test, expr_ret, Self=False, *expr_ret_args,
                 **expr_ret_kwargs):
        self._expr_test = expr_test
        self._expr_ret = expr_ret
        self._expr_ret_pass_self = Self
        self._expr_ret_args = expr_ret_args
        self._expr_ret_kwargs = expr_ret_kwargs

    def __call__(self, fn):
        """
        Executes a method call, if the test passed.

        :param instancemethod fn: decorated method
        """
        @wraps(fn)
        def wrapper(self_wr, *args, **kwargs):
            failed = False
            try:
                if not self._expr_test(self_wr):
                    failed = True
            except AttributeError, e:
                failed = True
            if failed:
                if self._expr_ret_pass_self:
                    return self._expr_ret(self_wr, *self._expr_ret_args,
                                          **self._expr_ret_kwargs)
                return self._expr_ret(*self._expr_ret_args,
                                      **self._expr_ret_kwargs)
            return fn(self_wr, *args, **kwargs)
        return wrapper


class lmi_return_val_if_fail(lmi_return_expr_if_fail):
    """
    Decorator, which returns a specified value and no method call is performed,
    if provided test expression is False; otherwise a method is called.

    :param expr_test: if the expression returns False, a method call is called
    :param rval: return value of the method, if the object attribute as
        expression failed

    Example of usage:

    .. code-block:: python

        class Foo:
            def __init__(self, member):
                self._member = member

            @lmi_return_val_if_fail(lambda obj: obj._member, False)
            def some_method(self):
                print "some_method called"
                return True

        f = Foo(None)
        f.some_method() == False

        f = Foo(True)
        f.some_method() == True
    """
    def __init__(self, expr_test, rval):
        lmi_return_expr_if_fail.__init__(self, expr_test, lambda: rval)


class lmi_return_if_fail(lmi_return_val_if_fail):
    """
    Decorator, which returns None and no method call is performed, if provided
    test expression is False; otherwise a method is called.

    :param expr_test: if the expression ``expr_test`` returns True, a method is
        called

    Example of usage:

    .. code-block:: python

        class Foo:
            def __init__(self, member):
                self._member = member

            @lmi_return_if_fail(lambda obj: obj._member)
            def some_method(self):
                print "some_method called"
                return True

        f = Foo(None)
        f.some_method() == None

        f = Foo(True)
        f.some_method() == True
    """
    def __init__(self, expr_test):
        lmi_return_val_if_fail.__init__(self, expr_test, None)


class lmi_possibly_deleted(lmi_return_expr_if_fail):
    """
    Decorator, which returns None, if provided test expression is True.

    :param expr_ret: callable or return value used, if ``expr_test`` fails
    :param expr_ret_args: expr_ret position arguments
    :param expr_ret_kwargs: expr_ret keyword arguments
    :param bool Self: flag, which specifies, if to pass ``self`` variable to
        the ``expr_ret``, if ``expr_test`` failed

    Example of usage:

    .. code-block:: python

        class Foo:
            def __init__(self, deleted):
                self._deleted = deleted

            @lmi_possibly_deleted(lambda obj: obj._member, lambda: False)
            def some_method(self):
                print "some_method called"
                return True

        f = Foo(None)
        f.some_method() == False

        f = Foo(True)
        f.some_method() == True
    """
    def __init__(self, expr_ret, Self=False, *expr_ret_args,
                 **expr_ret_kwargs):
        call_expr = expr_ret if callable(expr_ret) else lambda: expr_ret
        lmi_return_expr_if_fail.__init__(
            self, lambda obj: not obj._deleted, call_expr, Self=Self,
            *expr_ret_args, **expr_ret_kwargs)

    def __call__(self, fn):
        """
        Executes a method call, if the test passed. If the test expression
        returns False, specified return value in the constructor is returned.
        If the LMIShell does not dump exceptions, an appropriate exception is
        raised.

        :param instancemethod fn: decorated method
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        @wraps(fn)
        def wrapper(self_wr, *args, **kwargs):
            failed = False
            try:
                if not self._expr_test(self_wr):
                    failed = True
            except AttributeError, e:
                failed = True
            if failed:
                errorstr = "This instance has been deleted from a CIM broker"
                lmi_raise_or_dump_exception(LMIDeletedObjectError(errorstr))
                if self._expr_ret_pass_self:
                    return self._expr_ret(self_wr, *self._expr_ret_args,
                                          **self._expr_ret_kwargs)
                return self._expr_ret(*self._expr_ret_args,
                                      **self._expr_ret_kwargs)
            return fn(self_wr, *args, **kwargs)
        return wrapper


class lmi_class_fetch_lazy(object):
    """
    Decorator for :py:class:`.LMIClass`, which first fetches a wrapped
    :py:class:`wbem.CIMClass` object and then executes a wrapped method.

    :param bool full_fetch: True, if :py:class:`wbem.CIMClass` should include
        qualifiers and class origin. Default value is False.
    """
    def __init__(self, full_fetch=False):
        self._full_fetch = full_fetch

    def __call__(self, fn):
        """
        Fetches a :py:class:`wbem.CIMClass` if necessary and executes a wrapped
        method.

        :param instancemethod fn: decorated method
        """
        @wraps(fn)
        def wrapped(self_wr, *args, **kwargs):
            if not self_wr.is_fetched(self._full_fetch):
                self_wr.fetch(self._full_fetch)
            return fn(self_wr, *args, **kwargs)
        return wrapped

class lmi_instance_name_fetch_lazy(object):
    """
    Decorator for :py:class:`.LMIInstanceName`, which first fetches a wrapped
    :py:class:`wbem.CIMInstance` object and then executes a wrapped method.

    :param bool full_fetch: True, if :py:class:`wbem.CIMClass` should include
        qualifiers and class origin. Default value is False.
    """
    def __init__(self, full_fetch=False):
        self._full_fetch = full_fetch

    def __call__(self, fn):
        """
        Fetches a :py:class:`.LMIClass`, if necessary, and exeutes a wrapped
        method.

        :param instancemethod fn: decorated method
        """
        @wraps(fn)
        def wrapped(self_wr, *args, **kwargs):
            if self_wr._lmi_class is None:
                self_wr._lmi_class = lmi_wrap_cim_class(
                    self_wr._conn,
                    self_wr._cim_instance_name.classname,
                    self_wr._cim_instance_name.namespace)
                self_wr._lmi_class.fetch(self._full_fetch)
            return fn(self_wr, *args, **kwargs)
        return wrapped

# Default lmi_process_cim_exceptions error handler
def return_lmi_rval(rval, exc):
    errorstr = exc.args[1] if len(exc.args) > 1 else exc.args[0]
    return LMIReturnValue(rval=rval, errorstr=errorstr)


class lmi_process_cim_exceptions(object):
    """
    Decorator used for CIM-XML exception processing.

    :param rval: rval passed to :py:meth:`.LMIReturnValue.__init__`
    :param error_callable: callable used for processing
        :py:exc:`wbem.CIMError` and :py:exc:`.ConnectionError`

    **NOTE:** callables need to take 2 arguments: return value and error
    string.
    """

    def __init__(self, rval=None, error_callable=return_lmi_rval):
        self._rval = rval
        self._error_callable = error_callable

    def __call__(self, fn):
        """
        Execute a wrapped method.

        :param instancemethod fn: decorated method
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        @wraps(fn)
        def wrapped(*args, **kwargs):
            try:
                rval = fn(*args, **kwargs)
            except wbem.CIMError, e:
                lmi_raise_or_dump_exception(CIMError(e))
                return self._error_callable(self._rval, e)
            except wbem.ConnectionError, e:
                lmi_raise_or_dump_exception(ConnectionError(e))
                return self._error_callable(self._rval, e)
            except wbem.AuthError, e:
                # PyWBEM's AuthError does not have HTTP error code in the
                # exception args. We add the code by hand.
                e = ConnectionError(401, "Unauthorized")
                lmi_raise_or_dump_exception(e)
                return self._error_callable(self._rval, e)
            return rval
        return wrapped


class lmi_process_cim_exceptions_rval(lmi_process_cim_exceptions):
    """
    Decorator used for CIM-XML exception processing.

    :param rval: return value of a decorated method in case of exception
    """
    def __init__(self, rval=None):
        super(lmi_process_cim_exceptions_rval, self).__init__(
            rval, lambda r, e: r)


class lmi_process_wsman_exceptions(object):
    """
    Decorator used for wsman exception processing.

    :param rval: rval passed to :py:meth:`.LMIReturnValue.__init__`
    :param error_callable: callable used for processing
        :py:exc:`wbem.CIMError` and :py:exc:`.ConnectionError`

    **NOTE:** callables need to take 2 arguments: return value and error
    string.
    """
    def __init__(self, rval=None, error_callable=return_lmi_rval):
        self._rval = rval
        self._error_callable = error_callable

    def __call__(self, fn):
        """
        Execute a wrapped method.

        :param instancemethod fn: decorated method
        :raises: :py:exc:`.CIMError`
        """
        @wraps(fn)
        def wrapped(*args, **kwargs):
            try:
                rval = fn(*args, **kwargs)
            except wbem.CIMError, e:
                x = CIMError(e)
                lmi_raise_or_dump_exception(CIMError(e))
                return self._error_callable(self._rval, e)
            return rval
        return wrapped


class lmi_process_wsman_exceptions_rval(lmi_process_wsman_exceptions):
    """
    Decorator used for wsman exception processing.

    :param rval: return value of a decorated method in case of exception
    """
    def __init__(self, rval=None):
        super(lmi_process_wsman_exceptions_rval, self).__init__(
            rval, lambda r, e: r)



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Defines command class used to nest multiple commands under one.
"""

from docopt import docopt

from lmi.scripts.common import get_logger
from lmi.scripts.common.command import base
from lmi.scripts.common.command import meta

LOG = get_logger(__name__)

class LmiCommandMultiplexer(base.LmiBaseCommand):
    """
    Base class for node commands. It consumes just part of command line
    arguments and passes the remainder to one of its subcommands.

    Example usage: ::

        class MyCommand(LmiCommandMultiplexer):
            '''
            My command description.

            Usage: %(cmd)s mycommand (subcmd1 | subcmd2)
            '''
            COMMANDS = {'subcmd1' : Subcmd1, 'subcmd2' : Subcmd2}

    Where ``Subcmd1`` and ``Subcmd2`` are some other ``LmiBaseCommand``
    subclasses. Documentation string must be parseable with ``docopt``.

    Recognized properties:

        ``COMMANDS`` : ``dictionary``
            property will be translated to
            :py:meth:`LmiCommandMultiplexer.child_commands` class method by
            :py:class:`~.meta.MultiplexerMetaClass`.

    Using metaclass: :py:class:`.meta.MultiplexerMetaClass`.
    """
    __metaclass__ = meta.MultiplexerMetaClass

    @classmethod
    def child_commands(cls):
        """
        Abstract class method, that needs to be implemented in subclass.
        This is done by associated meta-class, when defining a command with
        assigned ``COMMANDS`` property.

        :returns: Dictionary of subcommand names with assigned command
            factories.
        :rtype: dictionary
        """
        raise NotImplementedError("child_commands must be implemented in"
                " a subclass")

    @classmethod
    def fallback_command(cls):
        """
        This is overriden by :py:class:`~.meta.MultiplexerMetaClass` when
        the ``FALLBACK_COMMAND`` gets processed.

        :returns: Command factory invoked for missing command on command line.
        :rtype: :py:class:`~.endpoint.LmiEndPointCommand`
        """
        return None

    @classmethod
    def is_end_point(cls):
        return False

    def run_subcommand(self, cmd_name, args):
        """
        Pass control to a subcommand identified by given name.

        :param string cmd_name: Name of direct subcommand, whose
            :py:meth:`~.base.LmiBaseCommand.run` method shall be invoked.
        :param list args: List of arguments for particular subcommand.
        :returns: Application exit code.
        :rtype: integer
        """
        if not isinstance(cmd_name, basestring):
            raise TypeError("cmd_name must be a string, not %s" %
                    repr(cmd_name))
        if not isinstance(args, (list, tuple)):
            raise TypeError("args must be a list, not %s" % repr(args))
        try:
            cmd_cls = self.child_commands()[cmd_name]
            cmd = cmd_cls(self.app, cmd_name, self)
        except KeyError:
            self.app.stderr.write(self.get_usage())
            LOG().critical('Unexpected command "%s".', cmd_name)
            return 1
        return cmd.run(args)

    def run(self, args):
        """
        Handle optional parameters, retrieve desired subcommand name and
        pass the remainder of arguments to it.

        :param list args: List of arguments with at least subcommand name.
        """
        if not isinstance(args, (list, tuple)):
            raise TypeError("args must be a list")
        full_args = self.get_cmd_name_parts(for_docopt=True) + args
        docopt_kwargs = {
            # check the --help ourselves (the default docopt behaviour checks
            # also for --version)
            'help' : False,
            # let's ignore options following first command for generated
            # usage string and when a height of this branch is > 2
            'options_first' : not self.has_own_usage()
                    or any(   not cmd.is_end_point()
                          for cmd in self.child_commands().values()
                          if  not args or args[0] not in self.child_commands()
                              or self.child_commands()[args[0]] is cmd)
        }
        options = docopt(self.get_usage(), full_args, **docopt_kwargs)
        if options.pop('--help', False) or (args and args[0] == '--help'):
            self.app.stdout.write(self.get_usage(proper=True))
            return 0
        if (   self.fallback_command() is not None
           and (not args or args[0] not in self.child_commands())):
            cmd_cls = self.fallback_command()
            cmd = cmd_cls(self.app, self.cmd_name, self.parent)
            return cmd.run(args)
        return self.run_subcommand(args[0], args[1:])



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module defining base command class for all possible commands of ``lmi``
meta-command.
"""

import abc
import re

# regular expression matching leading whitespaces not until the first line
# containing non-white-spaces
RE_LSPACES = re.compile(r'\A(\s*$.)*', re.DOTALL | re.MULTILINE)

#: Default formatting options overriden by options passed onc ommand-line and
#: set in configuration file.
DEFAULT_FORMATTER_OPTIONS = {
    'no_headings'    : False,
    'padding'        : 0,
    'human_friendly' : False,
}

class LmiBaseCommand(object):
    """
    Abstract base class for all commands handling command line arguments.
    Instances of this class are organized in a tree with root element being the
    ``lmi`` meta-command (if not running in interactive mode). Each such
    instance can have more child commands if its
    :py:meth:`LmiBaseCommand.is_multiplexer` method return ``True``. Each has
    one parent command except for the top level one, whose :py:attr:`parent`
    property returns ``None``.

    Set of commands is organized in a tree, where each command (except for the
    root) has its own parent. :py:meth:`is_end_point` method distinguishes
    leaves from nodes. The path from root command to the leaf is a sequence of
    commands passed to command line.

    There is also a special command called selector. Its :py:meth:`is_selector`
    method returns ``True``. It selects proper command that shall be passed all
    the arguments based on expression with profile requirements. It shares its
    name and parent with selected child.

    If the :py:meth:`LmiBaseCommand.has_own_usage` returns ``True``, the parent
    command won't process the whole command line and the remainder will be
    passed as a second argument to the :py:meth:`LmiBaseCommand.run` method.

    :param app: Main application object.
    :param string cmd_name: Name of command.
    :param parent: Parent command.
    :type parent: :py:class:`LmiBaseCommand`
    """

    __metaclass__ = abc.ABCMeta

    @classmethod
    def get_description(cls):
        """
        Return description for this command. This is usually a first line
        of documentation string of a class.

        :rtype: string
        """
        if cls.__doc__ is None:
            return ""
        return cls.__doc__.strip().split("\n", 1)[0]

    @classmethod
    def is_end_point(cls):
        """
        :returns: ``True``, if this command parses the rest of command line and
            can not have any child subcommands.
        :rtype: boolean
        """
        return True

    @classmethod
    def is_multiplexer(cls):
        """
        Is this command a multiplexer? Note that only one of
        :py:meth:`is_end_point`, :py:meth:`is_selector` and this method can
        evaluate to``True``.

        :returns: ``True`` if this command is not an end-point command and it's
            a multiplexer. It contains one or more subcommands. It consumes the
            first argument from command-line arguments and passes the rest to
            one of its subcommands.
        :rtype: boolean
        """
        return not cls.is_end_point()

    @classmethod
    def is_selector(cls):
        """
        Is this command a selector?

        :returns: ``True`` if this command is a subclass of
            :py:class:`lmi.scripts.common.command.select.LmiSelectCommand`.
        :rtype: boolean
        """
        return not cls.is_end_point() and not cls.is_multiplexer()

    @classmethod
    def has_own_usage(cls):
        """
        :returns: ``True``, if this command has its own usage string, which is
            returned by :py:meth:`LmiBaseCommand.get_description`. Otherwise
            the parent command must be queried.
        :rtype: boolean
        """
        return False

    @classmethod
    def child_commands(cls):
        """
        Abstract class method returning dictionary of child commands with
        structure: ::

            { "command-name" : cmd_factory, ... }

        Dictionary contains just a direct children (commands, which
        may immediately follow this particular command on command line).
        """
        raise NotImplementedError("child_commands() method must be overriden"
                " in a subclass")

    def __init__(self, app, cmd_name, parent=None):
        if not isinstance(cmd_name, basestring):
            raise TypeError('cmd_name must be a string')
        if parent is not None and not isinstance(parent, LmiBaseCommand):
            raise TypeError('parent must be an LmiBaseCommand instance')
        self._app = app
        self._cmd_name = cmd_name.strip()
        self._parent = parent

    @property
    def app(self):
        """ Return application object. """
        return self._app

    @property
    def parent(self):
        """ Return parent command. """
        return self._parent

    @property
    def cmd_name(self):
        """ Name of this subcommand as a single word. """
        return self._cmd_name

    @property
    def cmd_name_parts(self):
        """
        Convenience property calling :py:meth:`get_cmd_name_parts` to obtain
        command path as a list of all preceding command names.

        :rtype: list
        """
        return self.get_cmd_name_parts()

    @property
    def format_options(self):
        """
        Compose formatting options. Parent commands are queried for defaults. If
        command has no parent, default options will be taken from
        :py:attr:`DEFAULT_FORMATTER_OPTIONS` which are overriden by config
        settings.

        :returns: Arguments passed to formatter factory when formatter is
            for current command is constructed.
        :rtype: dictionary
        """
        if self.parent is None:
            options = DEFAULT_FORMATTER_OPTIONS.copy()
            options['no_headings'] = self.app.config.no_headings
            options['human_friendly'] = self.app.config.human_friendly
        else:
            options = self.parent.format_options
        return options

    @property
    def session(self):
        """
        :returns: Session object. Session for command and all of its children
            may be overriden with a call to :py:meth:`set_session_proxy`.
        :rtype: :py:class:`lmi.scripts.common.session.Session`
        """
        proxy = getattr(self, '_session_proxy', None)
        if proxy:
            return proxy
        if self.parent is not None:
            return self.parent.session
        return self.app.session

    def get_cmd_name_parts(self, all_parts=False, demand_own_usage=True,
            for_docopt=False):
        """
        Get name of this command as a list composed of names of all preceding
        commands since the top level one. When in interactive mode, only
        commands following the active one will be present.

        :param boolean full: Take no heed to the active command or interactive
            mode. Return all command names since top level node inclusive. This
            is overriden with *for_docopt* flag.
        :param boolean demand_own_usage: Wether to continue the upward
            traversal through command hieararchy past the active command until
            the command with its own usage is found. This is the default behaviour.
        :param boolean for_docopt: Docopt parser needs to be given arguments list
            without the first item compared to command names in usage string
            it receives. Thus this option causes skipping the first item that would
            be otherwise included.
        :returns: Command path. Returned list will always contain at least the
            name of this command.
        :rtype: list
        """
        parts = [self.cmd_name]
        cmd = self
        own_usage = cmd.has_own_usage()
        while (   cmd.parent is not None
              and (all_parts or self.app.active_command not in (cmd, cmd.parent))
              or  (demand_own_usage and not own_usage)):
            cmd = cmd.parent
            parts.append(cmd.cmd_name)
            own_usage = own_usage or cmd.has_own_usage()
        if for_docopt and parts:
            parts.pop()
        return list(reversed(parts))

    def get_usage(self, proper=False):
        """
        Get command usage. Return value of this function is used by docopt
        parser as usage string. Command tree is traversed upwards until command
        with defined usage string is found. End point commands (leaves) require
        manually written usage, so the first command in the sequence of parents
        with own usage string is obtained and its usage returned. For nodes
        missing own usage string this can be generated based on its
        subcommands.

        :param boolean proper: Says, whether the usage string written
            manually is required or not. It applies only to node (not a leaf)
            commands without its own usage string.
        """
        if self.is_end_point() or self.has_own_usage() or proper:
            # get proper (manually written) usage, also referred as *own*
            cmd = self
            while not cmd.has_own_usage() and cmd.parent is not None:
                cmd = cmd.parent
            if cmd.__doc__ is None:
                docstr = "Usage: %s\n" % " ".join(self.cmd_name_parts)
            else:
                docstr = ( ( cmd.__doc__.rstrip()
                           % {'cmd' : " ".join(cmd.cmd_name_parts)}
                           ))

                match = RE_LSPACES.match(docstr)
                if match:   # strip leading newlines
                    docstr = docstr[match.end(0):]

                match = re.match(r'^ +', docstr)
                if match:   # unindent help message
                    re_lspaces = re.compile(r'^ {%s}' % match.end(0))
                    docstr = "\n".join(re_lspaces.sub('', l)
                                for l in docstr.splitlines())
                docstr += "\n"

        else:
            # generate usage string from what is known, applies to nodes
            # without own usage
            hlp = []
            if self.get_description():
                hlp.append(self.get_description())
                hlp.append("")
            hlp.append("Usage:")
            hlp.append("    %s <command> [<args> ...]"
                    % " ".join(self.cmd_name_parts))
            hlp.append("")
            hlp.append("Commands:")
            cmd_max_len = max(len(c) for c in self.child_commands())
            for name, cmd in sorted(self.child_commands().items()):
                hlp.append(("  %%-%ds %%s" % cmd_max_len)
                        % (name, cmd.get_description()))
            docstr = "\n".join(hlp) + "\n"

        return docstr

    @abc.abstractmethod
    def run(self, args):
        """
        Handle the command line arguments. If this is not an end point
        command, it will pass the unhandled arguments to one of it's child
        commands. So the arguments are processed recursively by the instances
        of this class.

        :param list args: Arguments passed to the command line that were
            not yet parsed. It's the contents of ``sys.argv`` (if in
            non-interactive mode) from the current command on.
        :returns: Exit code of application. This maybe also be a boolean value
            or ``None``. ``None`` and ``True`` are treated as a success causing
            exit code to be 0.
        :rtype: integer
        """
        raise NotImplementedError("run method must be overriden in subclass")

    def set_session_proxy(self, session):
        """
        Allows to override session object. This is useful for especially for
        conditional commands (subclasses of
        :py:class:`~lmi.scripts.common.command.select.LmiSelectCommand`) that devide
        connections to groups satisfying particular expression. These groups
        are turned into session proxies containing just a subset of connections
        in global session object.

        :param session: Session object.
        """
        self._session_proxy = session



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Package with client-side python modules and command line utilities.
"""

import logging

from lmi.shell import LMINamespace
from lmi.shell import LMIExceptions
from lmi.scripts.common.configuration import Configuration
from lmi.scripts.common.lmi_logging import get_logger

LOG = get_logger(__name__)

def get_computer_system(ns):
    """
    Obtain an instance of ``CIM_ComputerSystem`` or its subclass. Preferred
    class name can be configured in configuration file. If such class does
    not exist, a base class (``CIM_ComputerSystem``) is enumerated instead.
    First feasible instance is cached and returned.

    :param ns: Namespace object where to look for computer system class.
    :type ns: :py:class:`lmi.shell.LMINamespace`
    :returns: Instance of ``CIM_ComputerSystem``.
    :rtype: :py:class:`lmi.shell.LMIInstance`.
    """
    if not isinstance(ns, LMINamespace):
        raise TypeError("ns must be an instance of LMINamespace")
    if not hasattr(get_computer_system, '_cs_cache'):
        get_computer_system._cs_cache = {}
    ns_path = ns.connection.uri + '/' + ns.name
    if not ns_path in get_computer_system._cs_cache:
        config = Configuration.get_instance()
        try:
            get_computer_system._cs_cache[ns_path] = cs = \
                    getattr(ns, config.system_class_name).first_instance()
        except LMIExceptions.LMIClassNotFound:
            LOG().warn('Failed to get instance of %s on host "%s"'
                    ' - falling back to CIM_ComputerSystem.',
                    config.system_class_name, ns.connection.uri)
            get_computer_system._cs_cache[ns_path] = cs = \
                    ns.CIM_ComputerSystem.first_instance_name()
        LOG().debug('Loaded instance of %s:%s for host "%s".',
            ns.name, cs.classname, ns.connection.uri)
    return get_computer_system._cs_cache[ns_path]



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

# Base class for classes, that are used for tab-completion.
# Used due to deprecated methods since python 2.2.


class LMIWrapperBaseObject(object):
    """
    Base class for all LMI wrapper classes, such as :py:class:`.LMINamespace`,
    :py:class:`.LMIClass`, :py:class:`.LMIInstanceName`,
    :py:class:`.LMIInstance`, :py:class:`.LMIMethod`.

    :param LMIConnection conn: connection object
    """
    def __init__(self, conn):
        super(LMIWrapperBaseObject, self).__init__()
        self._conn = conn

    @property
    def connection(self):
        """
        Property returning :class:`.LMIConnection` object.

        :returns: connection object
        :rtype: :py:class:`.LMIConnection`
        """
        return self._conn



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.


class LMIObjectFactory(object):
    """
    Object factory class. Used to avoid circular import dependencies between
    several LMI classes. The class implements a singleton design pattern.

    Example of usage:

    .. code-block:: python

        LMIObjectFactory().register(SomeClass)
        some_obj = LMIObjectFactory().SomeClass(*args, **kwargs)
    """
    # Singleton instance
    _instance = None

    def __getattr__(self, name):
        """
        Returns either a class member or a new instance of registered classes.

        :param string name: member name, which should be returned
        """
        if name in self._classes:
            return self._classes[name]
        raise AttributeError(name)

    def __new__(cls):
        """
        Creates a new :py:class:`.LMIObjectFactory` instance, if not created,
        and returns the singleton instance object.
        """
        if not cls._instance:
            cls._instance = super(LMIObjectFactory, cls).__new__(cls)
            cls._instance._classes = {}
        return cls._instance

    def register(self, reg_class):
        """
        Registers a class into the factory.
        """
        LMIObjectFactory._instance._classes[reg_class.__name__] = reg_class



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import pywsman
import types

from lmi.shell.compat import *

from lmi.shell.LMIDecorators import lmi_process_wsman_exceptions
from lmi.shell.LMIDecorators import lmi_process_wsman_exceptions_rval

from lmi.shell.LMIExceptions import CIMError
from lmi.shell.LMIExceptions import LMINotSupported

from lmi.shell.LMIReturnValue import LMIReturnValue

from lmi.shell.LMIUtil import lmi_instance_to_path
from lmi.shell.LMIUtil import lmi_parse_uri
from lmi.shell.LMIUtil import lmi_raise_or_dump_exception

SCHEMES = {
    "CIM": "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2",
    "PRS": "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2",
    "Win32": "http://schemas.microsoft.com/wbem/wsman/1/wmi",
    "OpenWBEM": "http://schema.openwbem.org/wbem/wscim/1/cim-schema/2",
    "Linux": "http://sblim.sf.net/wbem/wscim/1/cim-schema/2",
    "OMC": "http://schema.omc-project.org/wbem/wscim/1/cim-schema/2",
    "PG": "http://schema.openpegasus.org/wbem/wscim/1/cim-schema/2",
    "AMT": "http://intel.com/wbem/wscim/1/amt-schema/1",
    "IPS": "http://intel.com/wbem/wscim/1/ips-schema/1",
    "Sun": "http://schemas.sun.com/wbem/wscim/1/cim-schema/2",
    "Msvm": "http://schemas.microsoft.com/wbem/wsman/1/wmi",
    # XXX: Dell's iDrac uses this schema
    # "DCIM": "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2"
    "DCIM": "http://schemas.dell.com/wbem/wscim/1/cim-schema/2"
}


class LMIWSMANClient(object):
    """
    WS-MAN client.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be
        performed
    :param string password: user's password
    :param bool verify_server_cert: indicates, whether a server side
        certificate needs to be verified, if SSL used; default value is True
    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    """

    ASSOC_ASSOCIATORS, \
        ASSOC_REFERENCES, \
        RES_INSTANCE, \
        RES_INSTANCE_NAME = range(4)

    QUERY_LANG_CQL = "CQL"
    QUERY_LANG_WQL = "WQL"

    WSM_XML_FAULT, \
        WSM_XML_NO_RESPONSE, \
        WSM_XML_NO_FAULT = range(-1, -4, -1)

    def __init__(self, uri, username="", password="", interactive=False,
                 verify_server_cert=True, key_file=None, cert_file=None):
        if not uri.startswith("http://") and \
                not uri.startswith("https://"):
            uri = "https://" + uri
        self._uri = uri

        # Parse CIMOM URI
        scheme, hostname, port, path, username_, password_ = lmi_parse_uri(uri)
        if username_ is not None:
            username = username_
            password = password_

        # Create a WSMAN client
        self._cliconn = pywsman.Client(
            hostname, port, path, scheme, username, password)

        transport = self._cliconn.transport()
        if username:
            transport.set_auth_method(pywsman.BASIC_AUTH_STR)
        if cert_file:
            transport.set_cert(cert_file)
            transport.set_key(key_file)

        # X509 verification
        transport.set_verify_peer(verify_server_cert)
        transport.set_verify_host(verify_server_cert)

        # Add dummy auth_request_callback
        pywsman.Client.auth_request_callback = lambda cliconn: None

    @staticmethod
    def _get_filter(classname, inst_filter=None,
                    query_lang=QUERY_LANG_WQL):
        """
        Creates :py:class:`pywsman.Filter` out of inst_filter.

        :param str classname: string containing classname for the filter
        :param dict inst_filter: dictionary containing keys and values for the
            filter
        :param str query_lang: query language for the filter
        :rtype: :py:class:`pywsman.Filter`
        """
        if not inst_filter:
            return None
        query = "select * from %s where " % classname
        conds = []
        for k, v in inst_filter.iteritems():
            quotes = isinstance(v, basestring)
            cond = "%s =" % k
            if quotes:
                cond += " \"%s\"" % v
            else:
                cond += " %s" % v
            conds.append(cond)
        query += " and ".join(conds)

        filt = pywsman.Filter()
        if query_lang == LMIWSMANClient.QUERY_LANG_WQL:
            filt.wql(query)
        elif query_lang == LMIWSMANClient.QUERY_LANG_CQL:
            filt.cql(query)
        else:
            raise ValueError("Unknown query language '%s'" % query_lang)

        return filt

    @staticmethod
    def _query_to_scheme_prefix_classnames(query):
        """
        Transforms a query to WS-MAN prefix and list of class names.

        :param str query: query
        :rtype: tuple containing class name prefix, WS-MAN scheme and list of
            class names
        :raises: :py:exc:`ValueError`
        """
        tokens = query.split(" ")
        tokens_lower = [token.lower() for token in tokens]

        # CIM query languages *do not* define "data change" operations;
        # such as INSERT, UPDATE or DELETE.
        try:
            begin = tokens_lower.index("from") + 1
        except ValueError:
            raise ValueError("Wrong query")

        end = len(tokens)
        try:
            end = tokens_lower.index("where")
        except ValueError:
            pass

        classenames = []
        for token in tokens[begin:end]:
            classenames += [cls for cls in token.split(",") if cls]

        rprefix = ""
        rscheme = ""
        for cls in classenames:
            for prefix, scheme in SCHEMES.iteritems():
                if cls.startswith(prefix.lower()):
                    if rscheme and rscheme != scheme:
                        raise ValueError("Classes not from 1 scheme")
                    rprefix = prefix
                    rscheme = scheme

        return rprefix, rscheme, classenames

    @staticmethod
    def _uri_to_cls_ns(uri):
        """
        Transforms URI to class name and namespace.

        :param str uri: URI
        :rtype: tuple containing class name and namespace
        :raises: :py:exc:`ValueError`
        """
        for prefix, scheme in SCHEMES.iteritems():
            if not uri.startswith(scheme):
                continue
            try:
                cls, ns = uri[len(scheme) + 1:].rsplit("/", 1)[::-1]
                if cls == "*":
                    cls = None
            except ValueError:
                # Dell's iDrac doesn't include namespace in ResourceURI
                cls = uri[len(scheme) + 1:]
                ns = ""
            return cls, ns
        raise ValueError("Wrong URI format")

    @staticmethod
    def _uri_to_cls(uri):
        """
        Transforms URI to class name.

        :param str uri: URI
        :rtype: str
        """

        return LMIWSMANClient._uri_to_cls_ns(uri)[0]

    @staticmethod
    def _uri_to_ns(uri):
        """
        Transforms URI to namespace.

        :param str uri: URI
        :rtype: str
        """
        return LMIWSMANClient._uri_to_cls_ns(uri)[1]

    @staticmethod
    def _cls_ns_to_uri(classname, namespace, wildcard=False):
        """
        Creates URI from class name and namespace.

        :param str classname: class name
        :param str namespace: namespace
        :param bool wildcard: if class name is prefixed by "Win32" and wildcard
            is set to True, asterisk is used instead of the class name in
            resulting URI. Otherwise, the parameter is ignored.
        :rtype: str
        """
        for prefix, scheme in SCHEMES.iteritems():
            if not classname.startswith(prefix):
                continue
            if wildcard and prefix == "Win32":
                # Wildcards can be used with Microsoft WMI
                classname = "*"
            members = [
                m for m in [scheme, namespace, classname]
                if isinstance(m, basestring)
            ]
            return "/".join(members)
        raise ValueError("No scheme for class %s" % classname)

    @staticmethod
    def _xml_to_wbem_instance_name(epr, namespace, host=None):
        """
        Creates :py:class:`wbem.CIMInstanceName` from
        :py:class:`pywsman.XmlNode`.

        :param pywsman.XmlNode epr: XML node containing the instance name
        :param str namespace: string containing namespace of the instance name
        :param str host: string containing host name of the instance name
        :rtype: :py:class:`wbem.CIMInstanceName`
        """
        ref_parameters = epr.get("ReferenceParameters")
        uri = ref_parameters.get("ResourceURI")
        selectors = ref_parameters.get("SelectorSet")
        classname = LMIWSMANClient._uri_to_cls(str(uri))
        keybindings = wbem.NocaseDict()
        for selector in selectors:
            name = selector.attr_find(None, "Name").value()
            value = selector.child()
            if value and value.name() == "EndpointReference":
                value = LMIWSMANClient._xml_to_wbem_instance_name(
                    value, namespace, host)
            else:
                value = str(selector)
            keybindings[name] = value

        return wbem.CIMInstanceName(classname, keybindings, host, namespace)

    @staticmethod
    def _xml_to_wbem_property(prop, host=None):
        """
        Creates :py:class:`wbem.CIMProperty` from :py:class:`pywsman.XmlNode`.

        :param pywsman.XmlNode prop: XML node containing the property
        :param str host: string containing host name of a potential
            :py:class:`wbem.CIMInstanceName`
        :rtype: :py:class:`wbem.CIMProperty`
        """
        ref_parameters = prop.get("ReferenceParameters")
        if ref_parameters is not None:
            value = LMIWSMANClient._xml_to_wbem_instance_name(prop, host)
            t = "reference"
        else:
            value = str(prop)
            # For now, let's assume, all the properties are strings except
            # references. See above.
            t = "string"
        return wbem.CIMProperty(prop.name(), value, t)

    @staticmethod
    def _xml_to_wbem_instance(instance, namespace, host):
        """
        Creates :py:class:`wbem.CIMInstance` from :py:class:`pywsman.XmlNode`.

        :param pywsman.XmlNode instance: XML node containing the instance
        :param str namespace: string containing the namespace of the instance
        :param str host: string containing the host name of the instance
        :rtype: :py:class:`wbem.CIMInstance`
        """
        classname = instance.name()
        properties = wbem.NocaseDict()
        qualifiers = wbem.NocaseDict()
        property_list = []

        for prop in instance:
            properties[prop.name()] = LMIWSMANClient._xml_to_wbem_property(
                prop, host)
            property_list.append(prop.name())

        # Reconstruct wbem.CIMInstanceName
        # XXX: keybindings?
        keybindings = None
        path = wbem.CIMInstanceName(classname, keybindings, host, namespace)

        return wbem.CIMInstance(
            classname, properties, qualifiers, path, property_list)

    @staticmethod
    def _xml_check_response(doc):
        """
        Performs a XML response check and raises an exception, if necessary.

        :param pywsman.XmlDoc doc: XML doc containing WS-MAN response
        :raises: :py:exc:`.CIMError`
        """
        if doc is None:
            # We should get at least a XML response.
            raise wbem.CIMError(
                LMIWSMANClient.WSM_XML_NO_RESPONSE, "No response from CIMOM")
        elif doc.is_fault():
            fault = doc.fault()
            raise wbem.CIMError(LMIWSMANClient.WSM_XML_FAULT, fault.reason())

    @staticmethod
    def _xml_invoke_rval(method_doc, namespace, host):
        """
        Creates :py:class:`.LMIReturnValue` from :py:class:`pywsman.XmlNode`.

        :param pywsman.XmlNode method_doc: XML node containing the method
            result value and parameters
        :param str namespace: string containing namespace of a potential
            :py:class:`wbem.CIMInstanceName` in return parameters
        :param str host: string containing host name of potential
            :py:class:`wbem.CIMInstanceName` in return parameters
        :rtype: :py:class:`.LMIReturnValue`
        """
        rval = None
        rparams = {}
        for param in method_doc:
            name = param.name()
            value = param.child()
            if value and value.name() == "EndPointReference":
                value = LMIWSMANClient._xml_to_wbem_instance_name(
                    value, namespace, host)
            else:
                value = str(param)

            if name == "ReturnValue":
                rval = value
            else:
                rparams[name] = value

        return LMIReturnValue(rval=rval, rparams=rparams)

    @staticmethod
    def _path_to_xml(path):
        """
        Transforms :py:class:`wbem.CIMInstanceName` to XML string.

        :param wbem.CIMInstanceName path: instance name
        :rtype: string containing a XML node
        """
        epr = pywsman.EndPointReference(
            LMIWSMANClient._cls_ns_to_uri(path.classname, path.namespace),
            pywsman.WSA_TO_ANONYMOUS)
        for k, v in path.keybindings.iteritems():
            if isinstance(v, wbem.CIMInstanceName):
                value = LMIWSMANClient.path_to_xml(v)
            else:
                value = str(v)
            epr.add_selector(k, value)
        # Ugly, but needs to be done. PyWSMan escapes XML selector.
        # WSMan needs to encode whole EndPointReference into XML rather
        # than resource URI.
        return epr.to_xml().replace("&lt;", "<").replace("&gt;", ">")

    def connect(self):
        """
        Compatibility method present due to :py:class:`.LMICIMXMLClient`.
        """
        return LMIReturnValue(rval=True)

    def disconnect(self):
        """
        Compatibility method present due to :py:class:`.LMICIMXMLClient`.
        """
        pass

    def dummy(self):
        """
        Sends a "dummy" request to verify credentials.

        :returns: :py:class:`.LMIReturnValue` with rval set to True, if
            provided credentials are OK; False otherwise. If LMIShell uses
            exceptions, :py:exc:`.CIMError` will be raised.
        :raises: :py:exc:`.CIMError`
        """
        options = pywsman.ClientOptions()
        doc = self._cliconn.identify(options)

        try:
            self._xml_check_response(doc)
        except CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=False, errorstr=e.args[1])
        return LMIReturnValue(rval=True)

    def enumerate_iter_with_uri(self, uri, filt, options=None, limit=-1):
        """
        Enumerates instance (names).

        :param str uri: URI of the resource
        :param pywsman.Filter filt: filter for enumeration
        :param pywsman.ClientOptions options: options for enumeration
        :param int limit: enumeration limit
        :rtype: list containing :py:class:`wbem.CIMInstance` of
            :py:class:`wbem.CIMInstanceName`
        :raises: :py:exc:`.CIMError`
        """
        if options is None:
            options = pywsman.ClientOptions()

        namespace = self._uri_to_ns(uri)

        doc = self._cliconn.enumerate(options, filt, uri)
        self._xml_check_response(doc)

        context = doc.root().find(
            pywsman.XML_NS_ENUMERATION,
            "EnumerationContext")

        # Transformation function
        if options.get_flags() & pywsman.FLAG_ENUMERATION_ENUM_EPR:
            def xml_to_wbem_obj(doc):
                return self._xml_to_wbem_instance_name(
                    doc, namespace, self._cliconn.host())
        else:
            def xml_to_wbem_obj(doc):
                return self._xml_to_wbem_instance(
                    doc, namespace, self._cliconn.host())

        # Pull each Instance(Name)
        cnt = 0
        rval = []
        while context is not None and (cnt < limit or limit == -1):
            doc = self._cliconn.pull(options, None, uri, str(context))
            self._xml_check_response(doc)

            root = doc.root()
            context = root.find(
                pywsman.XML_NS_ENUMERATION,
                "EnumerationContext")
            items = root.find(pywsman.XML_NS_ENUMERATION, "Items")

            for item in items:
                rval.append(xml_to_wbem_obj(item))

            # Count the instances
            cnt += 1

        # Do we have all the instance(names)?
        if context is not None and cnt == limit:
            # Release the enumeration context from CIMOM.
            self._cliconn.release(options, uri, str(context))

        return rval

    def enumerate_iter(self, classname, namespace, filt, options=None,
                       limit=-1):
        """
        Enumerates instance (names).

        :param pywsman.Filter filt: filter for enumeration
        :param pywsman.ClientOptions options: options for enumeration
        :param int limit: enumeration limit
        :rtype: list containing :py:class:`wbem.CIMInstance` of
            :py:class:`wbem.CIMInstanceName`
        :raises: :py:exc:`.CIMError`
        """
        uri = self._cls_ns_to_uri(classname, namespace, bool(filt))
        return self.enumerate_iter_with_uri(uri, filt, options, limit)

    def enumerate(self, result_cls, classname, namespace=None,
                  inst_filter=None, limit=-1, **kwargs):
        """
        Enumerates instance (names).

        :param int result_cls: either :py:attr:`.LMIWSMANClient.RES_INSTANCE`
            or :py:attr:`.LMIWSMANClient.RES_INSTANCE_NAME`
        :param str classname: class name to enumerate
        :param str namespace: namespace where the class is located
        :param dict inst_filter: dictionary containing keys and values for the
            filter
        :param int limit: enumeration limit
        :param kwargs: keyword arguments used for inst_filter
        :rtype: list containing :py:class:`wbem.CIMInstance` of
            :py:class:`wbem.CIMInstanceName`
        :raises: :py:exc:`.CIMError`
        """
        filter_value = ""
        filter_key = "Name"
        if inst_filter is None:
            inst_filter = {}
        if "key" in kwargs:
            filter_key = kwargs["key"]
            kwargs.pop("key")
        elif "Key" in kwargs:
            filter_key = kwargs["Key"]
            kwargs.pop("Key")
        if "value" in kwargs:
            filter_value = kwargs["value"]
            kwargs.pop("value")
        if "Value" in kwargs:
            filter_value = kwargs["Value"]
            kwargs.pop("Value")
        if filter_value:
            inst_filter[filter_key] = filter_value

        options = pywsman.ClientOptions()
        if result_cls == LMIWSMANClient.RES_INSTANCE_NAME:
            options.set_flags(pywsman.FLAG_ENUMERATION_ENUM_EPR)
        filt = self._get_filter(
            classname, inst_filter, LMIWSMANClient.QUERY_LANG_WQL)

        return self.enumerate_iter(
            classname, namespace, filt, options, limit)

    def association(self, instance, relationship, result_cls, AssocClass=None,
                    ResultClass=None, Role=None, ResultRole=None, limit=-1):
        """
        Enumerates association instance (names).

        :param instance: object, for which the association objects will be
            enumerated. The object needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param relationship: :py:attr:`.LMIWSMANClient.ASSOC_ASSOCIATORS` or
            :py:attr:`.LMIWSMANClient.ASSOC_REFERENCES`
        :param result_cls: :py:attr:`.LMIWSMANClient.RES_INSTANCE` or
            :py:attr:`.LMIWSMANClient.RES_INSTANCE_NAME`
        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of names by mandating that each
            returned name identify an object that shall be associated to the
            source object through an instance of this class or one of its
            subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be either an instance of this class
            (or one of its subclasses) or be this class (or one of its
            subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of names by mandating that each returned name identify
            an object that shall be associated to the source object through an
            association in which the source object plays the specified role.
            That is, the name of the property in the association class that
            refers to the source object shall match the value of this
            parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be associated to the source object
            through an association in which the named returned object plays the
            specified role. That is, the name of the property in the
            association class that refers to the returned object shall match
            the value of this parameter.
        :param int limit: enumeration limit
        :returns: list of association objects
        """
        path = lmi_instance_to_path(instance)
        uri = self._cls_ns_to_uri(path.classname, path.namespace)

        # Create EndPointReference
        epr = pywsman.EndPointReference(uri, pywsman.WSA_TO_ANONYMOUS)
        for k, v in path.keybindings.iteritems():
            epr.add_selector(k, str(v))

        # Create filter
        filt = pywsman.Filter()
        if relationship == LMIWSMANClient.ASSOC_ASSOCIATORS:
            filt.associators(
                epr, AssocClass, ResultClass, Role, ResultRole, None, 0)
        else:
            filt.references(
                epr, AssocClass, ResultClass, Role, ResultRole, None, 0)

        options = pywsman.ClientOptions()
        if result_cls == LMIWSMANClient.RES_INSTANCE_NAME:
            options.set_flags(pywsman.FLAG_ENUMERATION_ENUM_EPR)

        # Enumerate association objects
        assoc = self.enumerate_iter(
            path.classname, path.namespace, filt, options, limit)

        return assoc

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    @lmi_process_wsman_exceptions()
    def get_instance_names(self, classname, namespace=None, inst_filter=None,
                           limit=-1, **kwargs):
        """
        Returns a list of :py:class:`wbem.CIMInstanceName` objects.

        :param string classname: class name
        :param string namespace: namespace name, where the instance names live
        :param dict inst_filter: dictionary containing filter values. The
            key corresponds to the primary key of the
            :py:class:`wbem.CIMInstanceName`; value contains the filtering
            value.
        :param int limit: enumeration limit
        :param dictionary kwargs: supported keyword arguments (these are
            **deprecated**)

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** (*string*) -- filtering value, see above

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` contains a
            list of :py:class:`wbem.CIMInstanceName` objects, if no error
            occurs; otherwise ``rval`` is set to None and ``errorstr`` contains
            appropriate error string
        :raises: :py:exc:`.CIMError`
        """
        inst_names_list = self.enumerate(
            LMIWSMANClient.RES_INSTANCE_NAME, classname, namespace,
            inst_filter, limit, **kwargs)
        return LMIReturnValue(rval=inst_names_list)

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    @lmi_process_wsman_exceptions()
    def get_instances(self, classname, namespace=None, inst_filter=None,
                      client_filtering=False, limit=-1, **kwargs):
        """
        Returns a list of :py:class:`wbem.CIMInstance` objects.

        :param string classname: class name
        :param string namespace: namespace, where the instances live
        :param dictionary inst_filter: dictionary containing filter values. The
            key corresponds to the primary key of the
            :py:class:`wbem.CIMInstanceName`; value contains the filtering
            value.
        :param bool client_filtering: if True, client-side filtering will be
            performed, otherwise the filtering will be done by a CIMOM. Default
            value is False.
        :param int limit: enumeration limit
        :param dictionary kwargs: supported keyword arguments (these are
            **deprecated**)

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** (*string*) -- filtering value, see above

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to a
            list of :py:class:`wbem.CIMIntance` objects, if no error occurs;
            otherwise ``rval`` is set to None and ``errorstr`` is set to
            corresponding error string.
        :raises: :py:exc:`.CIMError`
        """
        # We do not enumerate instances directly. First instance names are
        # pulled from CIMOM. This is done due to lack of keybindings
        # information in the pull response of the instance.
        inst_names_list = self.enumerate(
            LMIWSMANClient.RES_INSTANCE_NAME, classname, namespace,
            inst_filter, limit, **kwargs)

        # Get all the instances from instance names. Now we can construct
        # LMIInstance also with object path information.
        inst_list = []
        for inst_name in inst_names_list:
            inst, _, errorstr = self.get_instance(inst_name)
            if not inst:
                return LMIReturnValue(rval=[], errorstr=errorstr)
            inst_list.append(inst)

        # Client-side filtering present due to LMICIMXMLClient compatibility.
        if inst_list and inst_filter and client_filtering:
            inst_list_filtered = []
            for inst in inst_list:
                for filter_key, filter_value in inst_filter.iteritems():
                    if filter_key not in inst.properties or \
                            inst.properties[filter_key].value != filter_value:
                        break
                else:
                    inst_list_filtered.append(inst)
            inst_list = inst_list_filtered

        return LMIReturnValue(rval=inst_list)

    @lmi_process_wsman_exceptions()
    def get_instance(self, instance, LocalOnly=True, IncludeQualifiers=False,
                     IncludeClassOrigin=False, PropertyList=None):
        """
        Returns a :py:class:`wbem.CIMInstance` object.

        :param instance: path of the object, which is about to be retrieved.
            The object needs to be instance of following classes:

            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`wbem.CIMInstance`
            * :py:class:`.LMIInstanceName`
            * :py:class:`.LMIInstance`

        :param LocalOnly: unused
        :param IncludeQualifiers: unused
        :param IncludeClassOrigin: unused
        :param PropertyList: unused
        :returns: :py:class:`.LMIReturnValue` object, where ``rval`` is set to
            :py:class:`wbem.CIMInstance` object, if no error occurs; otherwise
            ``errorstr`` is set to corresponding error string
        :raises: :py:exc:`.CIMError`
        """
        path = lmi_instance_to_path(instance)

        options = pywsman.ClientOptions()
        for k, v in path.keybindings.iteritems():
            if isinstance(v, wbem.CIMInstanceName):
                value = self._path_to_xml(v)
            else:
                value = str(v)
            options.add_selector(k, value)
        uri = self._cls_ns_to_uri(path.classname, path.namespace)
        doc = self._cliconn.get(options, uri)
        self._xml_check_response(doc)

        inst = self._xml_to_wbem_instance(
            doc.body().child(), path.namespace, self._cliconn.host())

        # Adjust CIM Object Path by hand
        inst.path = path
        return LMIReturnValue(rval=inst)

    def get_class_names(self, *args, **kwargs):
        """
        Not supported.
        """
        lmi_raise_or_dump_exception(LMINotSupported("GetClassNames()"))
        return LMIReturnValue(rval=[], errorstr="GetClassName() not supported")

    def get_class(self, *args, **kwargs):
        """
        Not supported.
        """
        lmi_raise_or_dump_exception(LMINotSupported("GetClass()"))
        return LMIReturnValue(rval=None, errorstr="GetClass() not supported")

    def get_superclass(self, *args, **kwargs):
        """
        Not supported.
        """
        lmi_raise_or_dump_exception(LMINotSupported("GetSuperClass()"))
        return LMIReturnValue(
            rval=None, errorstr="GetSuperClass() not supported")

    @lmi_process_wsman_exceptions(-1)
    def call_method(self, instance, method, **params):
        """
        Executes a method within a given instance.

        :param instance: object, on which the method will be executed. The
            object needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string method: string containing a method name
        :param dictionary params: parameters passed to the method call
        :returns: :py:class:`.LMIReturnValue` object with rval set to return
            value of the method call, rparams set to returned parameters from
            the method call, if no error occurs; otherwise rval is set to -1
            and errorstr to appropriate error string
        :raises: :py:exc:`.CIMError`
        """
        path = lmi_instance_to_path(instance)

        options = pywsman.ClientOptions()
        for k, v in path.keybindings.iteritems():
            options.add_selector(k, str(v))
        uri = self._cls_ns_to_uri(path.classname, path.namespace)

        # Add method parameters
        for k, v in params.iteritems():
            options.add_property(k, str(v))

        # Invoke the method
        doc = self._cliconn.invoke(options, uri, method)
        self._xml_check_response(doc)

        method_doc = doc.body().get(method + "_OUTPUT")
        return self._xml_invoke_rval(
            method_doc, path.namespace, self._cliconn.host())

    @lmi_process_wsman_exceptions_rval([])
    def get_associator_names(self, instance, AssocClass=None, ResultClass=None,
                             Role=None, ResultRole=None, limit=-1):
        """
        Returns a list of associated :py:class:`wbem.CIMInstanceName` objects
        with an input instance.

        :param instance: for this object the list of associated
            :py:class:`wbem.CIMInstanceName` will be returned. The object needs
            to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of names by mandating that each
            returned name identify an object that shall be associated to the
            source object through an instance of this class or one of its
            subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be either an instance of this class
            (or one of its subclasses) or be this class (or one of its
            subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of names by mandating that each returned name identify
            an object that shall be associated to the source object through an
            association in which the source object plays the specified role.
            That is, the name of the property in the association class that
            refers to the source object shall match the value of this
            parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be associated to the source object
            through an association in which the named returned object plays the
            specified role. That is, the name of the property in the
            association class that refers to the returned object shall match
            the value of this parameter.
        :param int limit: enumeration limit
        :returns: list of associated :py:class:`wbem.CIMInstanceName` objects
            with an input instance, if no error occurs; otherwise en empty list
            is returned
        :raises: :py:exc:`.CIMError`
        """
        return self.association(
            instance, LMIWSMANClient.ASSOC_ASSOCIATORS,
            LMIWSMANClient.RES_INSTANCE_NAME, AssocClass, ResultClass, Role,
            ResultRole, limit)

    @lmi_process_wsman_exceptions_rval([])
    def get_associators(self, instance, AssocClass=None, ResultClass=None,
                        Role=None, ResultRole=None, IncludeQualifiers=False,
                        IncludeClassOrigin=False, PropertyList=None, limit=-1):
        """
        Returns a list of associated :py:class:`wbem.CIMInstance` objects with
        an input instance.

        :param instance: for this object the list of associated
            :py:class:`wbem.CIMInstance` objects will be returned. The object
            needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of objects by mandating that each
            returned object shall be associated to the source object through an
            instance of this class or one of its subclasses. Default value is
            None.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be either an instance of this class (or one of its
            subclasses) or be this class (or one of its subclasses). Default
            value is None.
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall be associated with the source object through an association
            in which the source object plays the specified role. That is, the
            name of the property in the association class that refers to the
            source object shall match the value of this parameter.  Default
            value is None.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be associated to the source object through an association in
            which the returned object plays the specified role. That is, the
            name of the property in the association class that refers to the
            returned object shall match the value of this parameter.  Default
            value is None.
        :param IncludeQualifiers: unused
        :param IncludeClassOrigin: unused
        :param PropertyList: unused
        :param int limit: enumeration limit
        :returns: list of associated :py:class:`wbem.CIMInstance` objects with
            an input instance, if no error occurs; otherwise an empty list is
            returned
        :raises: :py:exc:`.CIMError`
        """
        return self.association(
            instance, LMIWSMANClient.ASSOC_ASSOCIATORS,
            LMIWSMANClient.RES_INSTANCE, AssocClass,
            ResultClass, Role, ResultRole, limit)

    @lmi_process_wsman_exceptions_rval([])
    def get_reference_names(self, instance, ResultClass=None, Role=None,
                            limit=-1):
        """
        Returns a list of association :py:class:`wbem.CIMInstanceName` objects
        with an input instance.

        :param instance: for this object the association
            :py:class:`wbem.CIMInstanceName` objects will be returned. The
            object needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of object names by mandating that each returned
            Object Name identify an instance of this class (or one of its
            subclasses) or this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of object names by mandating that each returned object
            name shall identify an object that refers to the target instance
            through a property with a name that matches the value of this
            parameter.
        :param int limit: enumeration limit
        :returns: list of association :py:class:`wbem.CIMInstanceName` objects
            with an input instance, if no error occurs; otherwise an empty list
            is returned
        :raises: :py:exc:`.CIMError`
        """
        return self.association(
            instance, LMIWSMANClient.ASSOC_REFERENCES,
            LMIWSMANClient.RES_INSTANCE_NAME, None,
            ResultClass, Role, None, limit)

    @lmi_process_wsman_exceptions_rval([])
    def get_references(self, instance, ResultClass=None, Role=None,
                       IncludeQualifiers=False, IncludeClassOrigin=False,
                       PropertyList=None, limit=-1):
        """
        Returns a list of association :py:class:`wbem.CIMInstance` objects with
        an input instance.

        :param instance: for this object the association
            :py:class:`wbem.CIMInstances` objects will be returned. The
            object needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be an instance of this class (or one of its subclasses) or
            this class (or one of its subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall refer to the target object through a property with a name
            that matches the value of this parameter. Default value is None.
        :param IncludeQualifiers: unused
        :param IncludeClassOrigin: unused
        :param PropertyList: unused
        :param int limit: enumeration limit
        :returns: list of association :py:class:`wbem.CIMInstance` objects with
            an input instance, if no error occurs; otherwise an empty list is
            returned
        :raises: :py:exc:`.CIMError`
        """
        return self.association(
            instance, LMIWSMANClient.ASSOC_REFERENCES,
            LMIWSMANClient.RES_INSTANCE, None,
            ResultClass, Role, None, limit)

    def create_instance(self, *args, **kwargs):
        """
        Not supported.
        """
        lmi_raise_or_dump_exception(LMINotSupported("CreateInstance()"))
        return LMIReturnValue(
            rval=None, errorstr="CreateInstance() not supported")

    def modify_instance(self, *args, **kwargs):
        """
        Not supported.
        """
        lmi_raise_or_dump_exception(LMINotSupported("ModifyInstance()"))
        return LMIReturnValue(
            rval=False, errorstr="ModifyInstance() not supported")

    def delete_instance(self, *args, **kwargs):
        """
        Not supported.
        """
        lmi_raise_or_dump_exception(LMINotSupported("DeleteInstance()"))
        return LMIReturnValue(
            rval=False, errorstr="DeleteInstance() not supported")

    @lmi_process_wsman_exceptions()
    def exec_query(self, query_lang, query, namespace=wbem.DEFAULT_NAMESPACE):
        """
        Executes a query and returns a list of :py:class:`wbem.CIMInstance`
        objects.

        :param string query_lang: query language
        :param string query: query to execute
        :param string namespace: target namespace for the query
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to list
            of :py:class:`wbem.CIMInstance` objects, if no error occurs;
            otherwise ``rval`` is set to None and ``errorstr`` is set to
            corresponding error string
        :raises: :py:exc:`.CIMError`
        """
        # Parse CIM query and return WSMAN scheme, scheme prefix and classes
        # defined in the query.
        scheme, prefix, classes = self._query_to_scheme_prefix_classnames(
            query)

        # Grammar of CIM query language states, that there can be a list of
        # classes present in FROM statement. We support only one class.
        if len(classes) != 1:
            raise ValueError("Wrong query, too many classnames")
        classname = classes[0]

        filt = pywsman.Filter()
        if query_lang == LMIWSMANClient.QUERY_LANG_WQL:
            filt.wql(query)
        elif query_lang == LMIWSMANClient.QUERY_LANG_CQL:
            filt.cql(query)
        else:
            lmi_raise_or_dump_exception(
                ValueError("Unknown query language '%s'" % query_lang))

        uri = self._cls_ns_to_uri(classname, namespace, True)

        insts = self.enumerate_iter_with_uri(uri, filt)
        return LMIReturnValue(rval=insts)

    @property
    def username(self):
        """
        :returns: user name as a part of provided credentials
        :rtype: string
        """
        return self._cliconn.user()

    @property
    def uri(self):
        """
        :returns: URI of the CIMOM
        :rtype: string
        """
        return self._uri

    @property
    def hostname(self):
        """
        :returns: hostname of CIMOM
        :rtype: string
        """
        return self._cliconn.host()



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import argparse

from lmi.shell.LMIShellVersion import __version__


class LMIShellOptionsHelpWithVersionFormatter(argparse.HelpFormatter):
    """
    Helper class used for help message formatting.
    """
    def _format_usage(self, usage, actions, groups, prefix):
        """
        Returns a string with pre-formatted LMIShell usage. This method
        prepends LMIShell version to the usage message.
        """
        return argparse.HelpFormatter._format_usage(
            self, usage, actions, groups, "LMIShell v%s\n\n" % __version__)


class LMIShellOptionParser(argparse.ArgumentParser):
    """
    Helper class for CLI option parsing.
    """
    def error(self, msg):
        """
        Prints help message, error message and exits with erro code 2.
        """
        self.print_help()
        self.exit(2, "\n%s: error: %s\n" % (self.prog, msg))


class LMIShellOptions(object):
    """
    Class representing a LMIShell command line options. In the constructor, all
    command line options before a script name are passed to the LMIShell. First
    position argument belongs to the script and the rest of command line
    options is passed to the script run under the LMIShell.

    :param list argv: CLI arguments
    """
    _LOG_DEFAULT, \
        _LOG_VERBOSE, \
        _LOG_MORE_VERBOSE, \
        _LOG_QUIET = range(4)

    def __init__(self, argv):
        # Create a option parser
        parser = LMIShellOptionParser(
            usage="%(prog)s [options] script [script-options]",
            formatter_class=LMIShellOptionsHelpWithVersionFormatter)
        parser.add_argument(
            "-i", "--interact", dest="interact", action="store_true",
            default=False, help="inspect interactively after running a script")
        parser.add_argument(
            "-v", "--verbose", dest="verbose", action="store_true",
            default=False, help="print log messages to stderr")
        parser.add_argument(
            "-m", "--more-verbose", dest="more_verbose", action="store_true",
            default=False, help="print all log messages to stderr")
        parser.add_argument(
            "-q", "--quiet", dest="quiet", action="store_true", default=False,
            help="do not print any log messages to stderr")
        parser.add_argument(
            "-n", "--noverify", dest="verifycert", action="store_false",
            default=True, help="do not verify CIMOM SSL certificate")
        parser.add_argument(
            "--cwd-first-in-path", dest="cwd_first_in_path",
            action="store_true", default=False,
            help="prepend CWD in sys.path instead of appending it")

        # Split CLI arguments into LMIShell ones and script ones. LMIShell
        # arguments are those, which are before the first positional argument.
        # Other CLI arguments are passed to the interpreted script.
        try:
            dmap = map(lambda arg: not arg.startswith("-"), argv[1:])
            delimiter = dmap.index(True) + 1
        except ValueError, e:
            delimiter = len(argv)
        self._script_argv = argv[delimiter:]
        self._script_name = self._script_argv[0] if self._script_argv else ""

        # Parse CLI options
        shell_argv = argv[1:delimiter]
        options = parser.parse_args(shell_argv)

        # Store CLI options
        self._interact = options.interact
        self._verifycert = options.verifycert
        self._cwd_first_in_path = options.cwd_first_in_path
        self._log = LMIShellOptions._LOG_DEFAULT
        if options.verbose:
            self._log = LMIShellOptions._LOG_VERBOSE
        if options.more_verbose:
            self._log = LMIShellOptions._LOG_MORE_VERBOSE
        if options.quiet:
            self._log = LMIShellOptions._LOG_QUIET

        # Check for errors
        verbose_options = (
            options.verbose,
            options.more_verbose,
            options.quiet)
        if verbose_options.count(True) > 1:
            parser.error("Options -v, -m and -q are mutually exclusive")

    @property
    def interactive(self):
        """
        :returns: flag, which tells if the LMIShell should be initially run in
            the interactive mode
        :rtype: bool
        """
        return not self._script_argv

    @property
    def interact(self):
        """
        :returns: flag, which indicates, if the LMIShell should enter an
            interactive mode, after executing a provided script. The behavior
            is similar to python interpreter
        :rtype: bool
        """
        return self._interact

    @property
    def script_name(self):
        """
        :returns: script name, which is about to be run under the LMIShell
        :rtype: string
        """
        return self._script_name

    @property
    def script_argv(self):
        """
        :returns: list of command line arguments of the interpreted script
        """
        return self._script_argv

    @property
    def log(self):
        """
        :returns: log level
        :rtype: int

        Log level can be one of the following:

            * ``_LOG_DEFAULT``
            * ``_LOG_VERBOSE``
            * ``_LOG_MORE_VERBOSE``
            * ``_LOG_QUIET``
        """
        return self._log

    @property
    def verify_server_cert(self):
        """
        :returns: flag, which indicates, if LMIShell should verify server side
            certificate, if SSL used
        :rtype: bool
        """
        return self._verifycert

    @property
    def cwd_first_in_path(self):
        """
        :returns: True, if CWD should be prepended in sys.path; False if
            appended
        """
        return self._cwd_first_in_path



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Manager module for direct subcommands of lmi metacommand. Most of them are
loaded from entry_points of installed python eggs.
"""

import pkg_resources

from lmi.scripts.common import Configuration
from lmi.scripts.common import errors
from lmi.scripts.common import get_logger
from lmi.scripts.common.command import base
from lmi.scripts.common.command import util

LOG = get_logger(__name__)

class _CustomCommandWrapper(object):
    """
    Provide an interface mocking an entry_point object for custom commands
    added by lmi metacommand application.

    :param name: (``str``) Name of command.
    :param cmd_class: (``LmiBaseCommand``) Factory for custom commands.
    """

    def __init__(self, name, cmd_class):
        if not isinstance(name, basestring):
            raise TypeError("name must be a string")
        if not issubclass(cmd_class, base.LmiBaseCommand):
            raise TypeError("cmd_class must be a LmiBaseCommand")
        self._name = name
        self._cmd_class = cmd_class

    @property
    def name(self):
        """ Return command name. """
        return self._name

    def load(self):
        """ Return command class. """
        return self._cmd_class

class CommandManager(object):
    """
    Manager of direct subcommands of lmi metacommand. It manages commands
    registered with entry_points under particular namespace installed by
    python eggs. Custom commands may also be added.

    :param namespace: (``str``) Namespace, where commands are registered.
        For example ``lmi.scripts.cmd``.
    """

    def __init__(self, namespace=None):
        if namespace is not None and not isinstance(namespace, basestring):
            raise TypeError("namespace must be a string")
        if namespace is None:
            namespace = Configuration.get_instance().get_safe(
                    "Main", "CommandNamespace")
        self._namespace = namespace
        self._commands = {}
        self._load_commands()

    @property
    def command_names(self):
        """ Returns list of command names. """
        return self._commands.keys()

    def __len__(self):
        return len(self._commands)

    def __iter__(self):
        """ Yields command names. """
        return iter(self._commands)

    def __getitem__(self, cmd_name):
        """ Gets command factory for name. """
        return self.find_command(cmd_name)

    def _load_commands(self):
        """ Loads commands from entry points under provided namespace. """
        def _add_entry_point(epoint):
            """
            Convenience function taking an entry point, making some name
            checks and adding it to registered commands.
            """
            if not util.RE_COMMAND_NAME.match(epoint.name):
                LOG().error('Invalid command name: %s, ignoring.', epoint.name)
                return
            if epoint.name in self._commands:
                LOG().warn('Command "%s" already registered, ignoring.',
                        epoint.name)
            else:
                LOG().debug('Found registered command "%s".', epoint.name)
                self._commands[epoint.name] = epoint

        for entry_point in pkg_resources.iter_entry_points(self._namespace):
            if isinstance(entry_point, dict):
                for epoint in entry_point.values():
                    _add_entry_point(epoint)
            else:
                _add_entry_point(entry_point)


    def add_command(self, name, cmd_class):
        """
        Registers custom command. May be used for example for *help* command.

        :param name: (``str``) Name of command.
        :param cmd_class: (``LmiBaseCommand``) Factory for commands.
        """
        if not isinstance(name, basestring):
            raise TypeError("name must be a string")
        if not issubclass(cmd_class, base.LmiBaseCommand):
            raise TypeError("cmd_class must be a LmiBaseCommand")
        if not util.RE_COMMAND_NAME.match(name):
            raise errors.LmiCommandInvalidName(
                    cmd_class.__module__, cmd_class.__class__.__name__, name)
        if name in self._commands:
            LOG().warn('Command "%s" already managed, overwriting with "%s:%s".',
                    name, cmd_class.__module__, cmd_class.__name__)
        self._commands[name] = _CustomCommandWrapper(name, cmd_class)

    def find_command(self, cmd_name):
        """
        Loads a command associated with given name and returns it.

        :param cmd_name: (``str``) Name of command to load.
        :rtype: (``LmiBaseCommand``) Factory for commands.
        """
        try:
            return self._commands[cmd_name].load()
        except KeyError:
            raise errors.LmiCommandNotFound(cmd_name)
        except ImportError as err:
            LOG().debug('Failed to import command "%s".', cmd_name, exc_info=err)
            raise errors.LmiCommandImportError(
                    cmd_name, self._commands[cmd_name].module_name, err)

    def reload_commands(self, keep_custom=True):
        """
        Flushes all commands and reloads entry points.

        :param keep_custom: (``bool``) Custom commands -- not loaded from
            entry points -- are preserved.
        """
        if keep_custom:
            keep = dict((n, c) for n, c in self._commands.items()
                   if isinstance(c, _CustomCommandWrapper))
        else:
            keep = {}
        self._commands = {}
        self._load_commands()
        self._commands.update(keep)




# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys

from lmi.shell.compat import *

from lmi.shell.LMIObjectFactory import LMIObjectFactory
from lmi.shell.LMIBaseObject import LMIWrapperBaseObject
from lmi.shell.LMIDecorators import lmi_instance_name_fetch_lazy
from lmi.shell.LMIDecorators import lmi_possibly_deleted
from lmi.shell.LMIUtil import lmi_transform_to_lmi
from lmi.shell.LMIUtil import lmi_wrap_cim_method


class LMIInstanceName(LMIWrapperBaseObject):
    """
    LMI wrapper class representing :py:class:`wbem.CIMInstanceName`.

    :param LMIConnection conn: connection object
    :param CIMInstanceName cim_instance_name: wrapped object
    """
    def __init__(self, conn, cim_instance_name):
        if isinstance(cim_instance_name, LMIInstanceName):
            cim_instance_name = cim_instance_name.wrapped_object
        # We use __dict__ to avoid recursion potentially caused by
        # combo __setattr__ and __getattr__
        self.__dict__["_cim_instance_name"] = cim_instance_name
        self.__dict__["_deleted"] = False
        self.__dict__["_lmi_class"] = None
        super(LMIInstanceName, self).__init__(conn)

    def __cmp__(self, other):
        """
        :param LMIInstanceName other: :py:class:`.LMIInstanceName` object to
            compare
        :returns: negative number, if self < other; 0 if self == other or
            positive number, if self > other
        :rtype: int
        """
        if not isinstance(other, LMIInstanceName):
            return -1
        if self._deleted and not other._deleted:
            return -1
        elif not self._deleted and other._deleted:
            return 1
        return cmp(self._cim_instance_name, other._cim_instance_name)

    @lmi_possibly_deleted(False)
    def __contains__(self, key):
        """
        :param string key: key name, which will be tested for presence in
            keybindings
        :returns: True, if the specified key is present in keybindings, False
            otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return False. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.
        """
        return key in self._cim_instance_name

    @lmi_possibly_deleted(None)
    def __getattr__(self, name):
        """
        Returns class member or key property.

        :param string name: class member or key property name
        :returns: class member or key property
        :raises: :py:exc:`AttributeError`, :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.
        """
        if name in self._cim_instance_name:
            member = self._cim_instance_name[name]
            if isinstance(member, wbem.CIMInstanceName):
                member = lmi_transform_to_lmi(self._conn, member)
            return member
        elif not self._conn.is_wsman() and name in self.methods():
            return lmi_wrap_cim_method(self._conn, name, self)
        elif self._conn.is_wsman() and not name.startswith("_"):
            return lmi_wrap_cim_method(self._conn, name, self)
        raise AttributeError(name)

    @lmi_possibly_deleted(None)
    def __setattr__(self, name, value):
        """
        Assigns a new value to class member or key property.

        :param string name: class member or key property name
        :param value: new value to assign
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.
        """
        if name in self._cim_instance_name.keys():
            if isinstance(value, str):
                # Convert string value into unicode
                value = unicode(value, "utf-8")
            elif isinstance(
                    value,
                    (LMIObjectFactory().LMIInstanceName,
                     LMIObjectFactory().LMIInstance)):
                # Unpack wrapped LMI object
                value = value.wrapped_object
            self._cim_instance_name[name] = value
        else:
            self.__dict__[name] = value

    @lmi_possibly_deleted("")
    def __str__(self):
        """
        :returns: string containing object path
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty string.
        If the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.
        """
        return unicode(self).encode("utf-8")

    @lmi_possibly_deleted(u"")
    def __unicode__(self):
        """
        :returns: unicode string containing object path
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty unicode
        string. If the shell uses exceptions, :py:exc:`.LMIDeletedObjectError`
        will be raised.
        """
        return unicode(self._cim_instance_name)

    @lmi_possibly_deleted("")
    def __repr__(self):
        """
        :returns: pretty string for the object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty string.
        If the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.
        """
        return "%s(classname=\"%s\"...)" % (
            self.__class__.__name__,
            self.classname)

    @lmi_possibly_deleted([])
    def associator_names(self, AssocClass=None, ResultClass=None, Role=None,
                         ResultRole=None):
        """
        Returns a list of associated :py:class:`.LMIInstanceName` with this
        object.

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of names by mandating that each
            returned name identify an object that shall be associated to the
            source object through an instance of this class or one of its
            subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be either an instance of this class
            (or one of its subclasses) or be this class (or one of its
            subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of names by mandating that each returned name identify
            an object that shall be associated to the source object through an
            association in which the source object plays the specified role.
            That is, the name of the property in the association class that
            refers to the source object shall match the value of this
            parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be associated to the source object
            through an association in which the named returned object plays the
            specified role. That is, the name of the property in the
            association class that refers to the returned object shall match
            the value of this parameter.
        :returns: list of associated :py:class:`.LMIInstanceName` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty list. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        **Usage:** :ref:`associators_instance_names`.
        """
        assoc_names = self._conn.client.get_associator_names(
            self._cim_instance_name, AssocClass, ResultClass, Role, ResultRole)
        return lmi_transform_to_lmi(self._conn, assoc_names)

    @lmi_possibly_deleted(None)
    def first_associator_name(self, AssocClass=None, ResultClass=None,
                              Role=None, ResultRole=None):
        """
        Returns the first associated :py:class:`.LMIInstanceName` with this
        object.

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of names by mandating that each
            returned name identify an object that shall be associated to the
            source object through an instance of this class or one of its
            subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be either an instance of this class
            (or one of its subclasses) or be this class (or one of its
            subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of names by mandating that each returned name identify
            an object that shall be associated to the source object through an
            association in which the source object plays the specified role.
            That is, the name of the property in the association class that
            refers to the source object shall match the value of this
            parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be associated to the source object
            through an association in which the named returned object plays the
            specified role. That is, the name of the property in the
            association class that refers to the returned object shall match
            the value of this parameter.
        :returns: first associated :py:class:`.LMIInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        **Usage:** :ref:`associators_instance_names`.
        """
        assoc_names = self._conn.client.get_associator_names(
            self._cim_instance_name, AssocClass, ResultClass, Role, ResultRole,
            limit=1)
        if not assoc_names:
            return None
        return lmi_transform_to_lmi(self._conn, assoc_names[0])

    @lmi_possibly_deleted([])
    def associators(self, AssocClass=None, ResultClass=None, Role=None,
                    ResultRole=None, IncludeQualifiers=False,
                    IncludeClassOrigin=False, PropertyList=None):
        """
        Returns a list of associated :py:class:`.LMIInstance` objects with this
        instance.

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of objects by mandating that each
            returned object shall be associated to the source object through an
            instance of this class or one of its subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be either an instance of this class (or one of its
            subclasses) or be this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall be associated with the source object through an association
            in which the source object plays the specified role. That is, the
            name of the property in the association class that refers to the
            source object shall match the value of this parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be associated to the source object through an association in
            which the returned object plays the specified role. That is, the
            name of the property in the association class that refers to the
            returned object shall match the value of this parameter.
        :param bool IncludeQualifiers: bool flag indicating, if all qualifiers
            for each object (including qualifiers on the object and on any
            returned properties) shall be included as ``<QUALIFIER>`` elements
            in the response.
        :param bool IncludeClassOrigin: bool flag indicating, if the
            ``CLASSORIGIN`` attribute shall be present on all appropriate
            elements in each returned object.
        :param list PropertyList: if not None, the members of the array define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If
            *PropertyList* is an empty list, no properties are included in each
            returned object.  If it is None, no additional filtering is
            defined.
        :returns: list of associated :py:class:`.LMIInstance` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty list. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        **Usage:** :ref:`associators_instances`.
        """
        associators = self._conn.client.get_associators(
            self._cim_instance_name, AssocClass, ResultClass, Role, ResultRole,
            IncludeQualifiers, IncludeClassOrigin, PropertyList)
        return lmi_transform_to_lmi(self._conn, associators)

    @lmi_possibly_deleted(None)
    def first_associator(self, AssocClass=None, ResultClass=None, Role=None,
                         ResultRole=None, IncludeQualifiers=False,
                         IncludeClassOrigin=False, PropertyList=None):
        """
        Returns the first associated :py:class:`.LMIInstance` with this object.

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of objects by mandating that each
            returned object shall be associated to the source object through an
            instance of this class or one of its subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be either an instance of this class (or one of its
            subclasses) or be this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall be associated with the source object through an association
            in which the source object plays the specified role. That is, the
            name of the property in the association class that refers to the
            source object shall match the value of this parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be associated to the source object through an association in
            which the returned object plays the specified role. That is, the
            name of the property in the association class that refers to the
            returned object shall match the value of this parameter.
        :param bool IncludeQualifiers: bool flag indicating, if all qualifiers
            for each object (including qualifiers on the object and on any
            returned properties) shall be included as ``<QUALIFIER>`` elements
            in the response.
        :param bool IncludeClassOrigin: bool flag indicating, if the
            ``CLASSORIGIN`` attribute shall be present on all appropriate
            elements in each returned object.
        :param list PropertyList: if not None, the members of the array define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If PropertyList
            is an empty list, no properties are included in each returned
            object. If it is None, no additional filtering is defined.
        :returns: first associated :py:class:`.LMIInstance`
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        **Usage:** :ref:`associators_instances`.
        """
        associators = self._conn.client.get_associators(
            self._cim_instance_name, AssocClass, ResultClass, Role, ResultRole,
            IncludeQualifiers, IncludeClassOrigin, PropertyList, limit=1)
        if not associators:
            return None
        return lmi_transform_to_lmi(self._conn, associators[0])

    @lmi_possibly_deleted([])
    def reference_names(self, ResultClass=None, Role=None):
        """
        Returns a list of association :py:class:`.LMIInstanceName` objects with
        this object.

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of object names by mandating that each returned
            Object Name identify an instance of this class (or one of its
            subclasses) or this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of object names by mandating that each returned object
            name shall identify an object that refers to the target instance
            through a property with a name that matches the value of this
            parameter.
        :returns: list of association :py:class:`.LMIInstanceName` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty list. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        **Usage:** :ref:`references_instance_names`.
        """
        reference_names = self._conn.client.get_reference_names(
            self._cim_instance_name, ResultClass, Role)
        return lmi_transform_to_lmi(self._conn, reference_names)

    @lmi_possibly_deleted(None)
    def first_reference_name(self, ResultClass=None, Role=None):
        """
        Returns the first association :py:class:`.LMIInstanceName` with this
        object.

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of object names by mandating that each returned
            Object Name identify an instance of this class (or one of its
            subclasses) or this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of object names by mandating that each returned object
            name shall identify an object that refers to the target instance
            through a property with a name that matches the value of this
            parameter.
        :returns: first association :py:class:`.LMIInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        **Usage:** :ref:`references_instance_names`.
        """
        reference_names = self._conn.client.get_reference_names(
            self._cim_instance_name, ResultClass, Role, limit=1)
        if not reference_names:
            return None
        return lmi_transform_to_lmi(self._conn, reference_names[0])

    @lmi_possibly_deleted([])
    def references(self, ResultClass=None, Role=None, IncludeQualifiers=False,
                   IncludeClassOrigin=False, PropertyList=None):
        """
        Returns a list of association :py:class:`.LMIInstance` objects with
        this object.

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be an instance of this class (or one of its subclasses) or
            this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall refer to the target object through a property with a name
            that matches the value of this parameter.
        :param bool IncludeQualifiers: flag indicating, if all qualifiers for
            each object (including qualifiers on the object and on any returned
            properties) shall be included as ``<QUALIFIER>`` elements in the
            response.
        :param bool IncludeClassOrigin: flag indicating, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in each
            returned object.
        :param list PropertyList: if not None, the members of the list define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If PropertyList
            is an empty list, no properties are included in each returned
            object. If PropertyList is None, no additional filtering is
            defined.
        :returns: list of association :py:class:`.LMIInstance` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty list. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        **Usage:** :ref:`references_instances`.
        """
        references = self._conn.client.get_references(
            self._cim_instance_name, ResultClass, Role, IncludeQualifiers,
            IncludeClassOrigin, PropertyList)
        return lmi_transform_to_lmi(self._conn, references)

    @lmi_possibly_deleted(None)
    def first_reference(self, ResultClass=None, Role=None,
                        IncludeQualifiers=False, IncludeClassOrigin=False,
                        PropertyList=None):
        """
        Returns the first association :py:class:`.LMIInstance` with this
        object.

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be an instance of this class (or one of its subclasses) or
            this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall refer to the target object through a property with a name
            that matches the value of this parameter.
        :param bool IncludeQualifiers: flag indicating, if all qualifiers for
            each object (including qualifiers on the object and on any returned
            properties) shall be included as ``<QUALIFIER>`` elements in the
            response.
        :param bool IncludeClassOrigin: flag indicating, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in each
            returned object.
        :param list PropertyList: if not None, the members of the list define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If PropertyList
            is an empty list, no properties are included in each returned
            object. If PropertyList is None, no additional filtering is
            defined.
        :returns: first association :py:class:`.LMIInstance` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        **Usage:** :ref:`references_instances`.
        """
        references = self._conn.client.get_references(
            self._cim_instance_name, ResultClass, Role, IncludeQualifiers,
            IncludeClassOrigin, PropertyList, limit=1)
        if not references:
            return None
        return lmi_transform_to_lmi(self._conn, references[0])

    def copy(self):
        """
        :returns: copy of itself
        """
        return lmi_transform_to_lmi(self._conn, self._cim_instance_name.copy())

    @lmi_possibly_deleted(True)
    def delete(self):
        """
        Deletes the instance defined by this object path from the CIMOM.

        :returns: True, if the instance is deleted; False otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return True. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        **Usage:** :ref:`instance_names_delete`.
        """
        rval, _, _ = self._conn.client.delete_instance(
            self._cim_instance_name)
        self._deleted = rval
        return self._deleted

    @lmi_possibly_deleted(None)
    def to_instance(self):
        """
        Creates a new :py:class:`.LMIInstance` object from
        :py:class:`.LMIInstanceName`.

        :returns: :py:class:`.LMIInstance` object if the object was retrieved
            successfully; None otherwise.
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        **Usage:** :ref:`instance_names_conversion`.
        """
        instance, _, _ = self._conn.client.get_instance(
            self._cim_instance_name, LocalOnly=False)
        if not instance:
            return None
        return lmi_transform_to_lmi(self._conn, instance)

    @lmi_possibly_deleted([])
    @lmi_instance_name_fetch_lazy()
    def methods(self):
        """
        Returns a list of :py:class:`wbem.CIMInstance` methods' names.

        :returns: list of :py:class:`wbem.CIMInstance` methods' names
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty list. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        **Usage:** :ref:`instances_methods`.
        """
        return self._lmi_class.methods()

    @lmi_possibly_deleted(None)
    @lmi_instance_name_fetch_lazy()
    def print_methods(self):
        """
        Prints out the list of :py:class:`wbem.CIMInstance` methods' names.

        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        **Usage:** :ref:`instances_methods`.
        """
        self._lmi_class.print_methods()

    @lmi_possibly_deleted([])
    def key_properties(self):
        """
        :returns: list of strings of key properties
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty list. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        **Usage:** :ref:`instance_names_key_properties`.
        """
        return self._cim_instance_name.keys()

    @lmi_possibly_deleted(None)
    def print_key_properties(self):
        """
        Prints out the list of key properties.

        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        **Usage:** :ref:`instance_names_key_properties`.
        """
        for name, key_prop in self._cim_instance_name.iteritems():
            value = str(key_prop)
            if isinstance(key_prop, basestring):
                value = "'%s'" % value
            sys.stdout.write("%s = %s\n" % (name, value))

    @lmi_possibly_deleted({})
    def key_properties_dict(self):
        """
        :returns: dictionary with key properties and corresponding values
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty
        dictionary. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.
        """
        return self._cim_instance_name.keybindings.copy()

    @lmi_possibly_deleted(None)
    def key_property_value(self, prop_name):
        """
        :param string prop_name: key property name
        :returns: key property value
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.
        """
        return getattr(self, prop_name)

    @property
    @lmi_possibly_deleted("")
    def classname(self):
        """
        :returns: class name
        :rtype: string
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty string.
        If the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.
        """
        return self._cim_instance_name.classname

    @property
    @lmi_possibly_deleted("")
    def namespace(self):
        """
        :returns: namespace name
        :rtype: string
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty string.
        If the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.
        """
        return self._cim_instance_name.namespace

    @property
    @lmi_possibly_deleted("")
    def hostname(self):
        """
        :returns: host name
        :rtype: string
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return an empty string.
        If the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.
        """
        return self._cim_instance_name.host

    @property
    def is_deleted(self):
        """
        :returns: True, if the instance was deleted from the CIMOM; False
            otherwise
        """
        return self._deleted

    @property
    @lmi_possibly_deleted(None)
    def wrapped_object(self):
        """
        :returns: wrapped :py:class:`wbem.CIMInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **NOTE:** If the method :py:meth:`.LMIInstanceName.delete` was called,
        this method will not execute its code and will return None. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.
        """
        return self._cim_instance_name
